from typing import Union
from pathlib import Path
import numpy as np
import torch
from torch import nn

from tqdm.autonotebook import tqdm

from xmen.reranking import Reranker
from xmen.reranking.scored_cross_encoder import ScoredInputExample, ScoredCrossEncoder
from xmen.reranking.ranking_util import get_flat_candidate_ds

from sentence_transformers.cross_encoder import CrossEncoder

from transformers.trainer_utils import set_seed

from datasets import DatasetDict
from xmen.data import IndexedDatasetDict, IndexedDataset, Deduplicator

import logging
from sentence_transformers import LoggingHandler

logger = logging.getLogger(__name__)

from typing import Union, List


def flat_ds_to_cross_enc_dataset(flat_candidate_ds, doc_index, context_length, mask_mention, encode_sem_type, use_nil):
    """
    Convert a flat candidate dataset to a cross-encoding dataset.

    Args:
    - flat_candidate_ds (list): A list of dictionaries where each dictionary represents a single training example. The keys
                                of each dictionary are "context_left", "mention", "context_right", "candidates", "scores",
                                "synonyms", and "label".
    - doc_index (list): A list of indices for each document in the dataset.
    - context_length (int): The number of tokens to include in the context before and after the mention. If set to 0, the
                            context will not be truncated.
    - mask_mention (bool): Whether or not to mask the mention in the generated mention-candidate pairs.

    Returns:
    - res (list): A list of batches, where each batch is a list of ScoredInputExamples. Each ScoredInputExample represents a
                  single mention-candidate pair with a label and a score.
    - res_index (list): A list of indices corresponding to the batches in res.
    """
    res = []
    res_index = []
    for doc, idx in zip(tqdm(flat_candidate_ds), doc_index):
        l_ctx, m, r_ctx = doc["context_left"], doc["mention"], doc["context_right"]
        mention = f"{l_ctx[-context_length:] if context_length else ''} [START] {m if not mask_mention else '[MASK]'} [END] {r_ctx[:context_length] if context_length else ''}"
        batch = []
        match_found = False
        for c, score, syns, semtype in zip(doc["candidates"], doc["scores"], doc["synonyms"], doc["types"]):
            candidate = syns[0] + " [TITLE] " + " [SEP] ".join(syns[1:])
            if encode_sem_type:
                candidate = ",".join(semtype) + " [TYPE] " + candidate
            label = 1 if ("label" in doc and c in doc["label"]) else 0
            if label == 1:
                match_found = True
            batch.append(ScoredInputExample(texts=[mention, candidate], label=label, score=score))
        if batch:
            if use_nil:
                label = 1 if not match_found else 0
                batch[-1] = ScoredInputExample(texts=[mention, "[NIL]"], label=label, score=label, nil=True)
            res.append(batch)
            res_index.append(idx)
    return res, res_index


def create_cross_enc_dataset(
    candidate_ds,
    ground_truth,
    kb,
    context_length: int,
    expand_abbreviations: bool,
    encode_sem_type: bool,
    masking: bool,
    use_nil: bool,
):
    """
    Create a cross-encoding dataset from a candidate dataset.

    Args:
    - candidate_ds (list): A list of dictionaries where each dictionary represents a single candidate. The keys of each
                           dictionary are "entity", "context", and "mention".
    - ground_truth (dict): A dictionary where the keys are mentions and the values are lists of entities that are valid
                           candidates for that mention.
    - kb (KnowledgeBase): A KnowledgeBase object that contains information about valid entity types and relations between
                          entities.
    - context_length (int): The number of tokens to include in the context before and after the mention. If set to 0, the
                            context will not be truncated.
    - expand_abbreviations (bool): Whether or not to expand abbreviations in the context before and after the mention.
    - masking (bool): Whether or not to mask the mention in the generated mention-candidate pairs.

    Returns:
    - res (list): A list of batches, where each batch is a list of ScoredInputExamples. Each ScoredInputExample represents a
                  single mention-candidate pair with a label and a score.
    - res_index (list): A list of indices corresponding to the batches in res.
    """
    flat_candidate_ds, doc_index = get_flat_candidate_ds(
        candidate_ds, ground_truth, expand_abbreviations=expand_abbreviations, kb=kb
    )
    return flat_ds_to_cross_enc_dataset(
        flat_candidate_ds,
        doc_index,
        context_length,
        mask_mention=masking,
        encode_sem_type=encode_sem_type,
        use_nil=use_nil,
    )


def _cross_encoder_predict(cross_encoder, cross_enc_dataset, show_progress_bar, convert_to_numpy):
    """
    Generate cross-encoder predictions for a cross-encoder and a cross-encoding dataset.

    Args:
    - cross_encoder (CrossEncoder): A CrossEncoder object that will be used to generate predictions.
    - cross_enc_dataset (list): A list of batches, where each batch is a list of ScoredInputExamples. Each ScoredInputExample
                                represents a single mention-candidate pair with a label and a score.
    - show_progress_bar (bool): Whether or not to display a progress bar while generating predictions. If None, the default
                                behavior is to show the progress bar if the logger's effective level is set to INFO or
                                DEBUG.
    - convert_to_numpy (bool): Whether or not to convert the generated predictions to numpy arrays.

    Returns:
    - pred_scores (list): A list of tensors, where each tensor represents the predicted scores for a single mention-candidate
                          pair.
    """
    inp_dataloader = torch.utils.data.DataLoader(
        BatchedCrossEncoderDataset(cross_enc_dataset), num_workers=0, batch_size=None
    )
    inp_dataloader.collate_fn = cross_encoder.smart_batching_collate

    if show_progress_bar is None:
        show_progress_bar = logger.getEffectiveLevel() == logging.INFO or logger.getEffectiveLevel() == logging.DEBUG

    iterator = inp_dataloader
    if show_progress_bar:
        iterator = tqdm(inp_dataloader, desc="Batches")

    pred_scores = []
    cross_encoder.model.eval()
    # model.to(self._target_device)
    with torch.no_grad():
        for features, _ in iterator:
            model_predictions = cross_encoder.model(**features, return_dict=True)
            logits = model_predictions.logits
            logits = torch.nn.functional.softmax(logits.reshape(-1), dim=0)
            pred_scores.append(logits)

    if convert_to_numpy:
        pred_scores = [score.cpu().detach().numpy() for score in pred_scores]

    return pred_scores


def rerank(doc, index, doc_idx, ce_dataset, ranking, allow_nil):
    """
    Re-ranks entities in a given document based on their normalized scores.

    Args:
    - doc (dict): A dictionary containing the document to be re-ranked.
    - index (int): The index of the document.
    - doc_idx (list): A list of indices indicating the position of entities in the document.
    - ce_dataset (list): A list of ScoredInputExamples representing the mention-candidate pairs in the document.
    - ranking (list): A list of scores to rank the entities.

    Raises:
    - AssertionError: If an entity's normalized score has a different length than its corresponding ranking score.

    Returns:
    - A dictionary containing the re-ranked entities in the given document.
    """
    entities = []
    for ei, e in enumerate(doc["entities"]):
        mask = (np.array(doc_idx) == [index, ei]).all(axis=1)
        ranking_idx = mask.argmax()
        if mask[ranking_idx] == False:
            assert len(e["normalized"]) == 0
        else:
            candidates = ce_dataset[ranking_idx]
            rank = ranking[ranking_idx]
            assert len(e["normalized"]) == len(rank), (len(e["normalized"]), len(rank))
            for n, r, cand in zip(e["normalized"], rank, candidates):
                n["score"] = r
                if cand.nil:
                    if allow_nil:
                        n["db_id"] = None
                    else:
                        n["score"] = 0.0
            e["normalized"].sort(key=lambda k: k["score"], reverse=True)
            if allow_nil and not e["normalized"][0]["db_id"]:
                e["normalized"] = []
        entities.append(e)
    return {"entities": entities}


class CrossEncoderTrainingArgs:
    """
    A class to store arguments for training a CrossEncoder model.

    Args:
    - args (dict): A dictionary containing the arguments for training.

    Attributes:
    - args (dict): A dictionary containing the arguments for training, which can also be set individually.
    """

    def __init__(self, args: dict):
        self.args = args

    def __init__(
        self,
        num_train_epochs: int,
        model_name: str = "bert-base-multilingual-cased",
        fp16: bool = True,
        label_smoothing: bool = False,
        rank_regularization: bool = 1.0,
        train_layers: list = None,
        softmax_loss: bool = True,
        random_seed: int = 42,
        learning_rate: float = 2e-5,
    ):
        self.args = {}
        self.args["model_name"] = model_name
        self.args["num_train_epochs"] = num_train_epochs
        self.args["fp16"] = fp16
        self.args["label_smoothing"] = label_smoothing
        self.args["rank_regularization"] = rank_regularization
        self.args["train_layers"] = train_layers
        self.args["softmax_loss"] = softmax_loss
        self.args["random_seed"] = random_seed
        self.args["learning_rate"] = learning_rate

    def __getitem__(self, key):
        return self.args[key]

    def get(self, key, default_val):
        return self.args.get(key, default_val)


class CrossEncoderReranker(Reranker):
    """
    Reranker that uses a cross-encoder to score a set of candidate passages against a query.
    Inherits from the abstract class Reranker.
    """

    def __init__(self, model=None):
        self.model = model

    @staticmethod
    def load(checkpoint, device, max_length=512):
        """
        Loads a pre-trained model from a checkpoint and returns a new instance of CrossEncoderReranker.

        Args:
        - checkpoint: The path to the checkpoint file to load.
        - device: The device to load the model onto.
        - max_length: Maximum input length

        Returns:
        - new instance of CrossEncoderReranker.
        """
        model = CrossEncoder(checkpoint)
        model._target_device = torch.device(device)
        model.model.to(torch.device(device))
        if not model.max_length:
            model.max_length = max_length
        return CrossEncoderReranker(model)

    @staticmethod
    def prepare_data(
        candidates,
        ground_truth,
        kb,
        context_length: int = 128,
        expand_abbreviations: bool = False,
        encode_sem_type: bool = False,
        masking: bool = False,
        use_nil: bool = True,
        **kwargs,
    ):
        """
        Prepares the data for training or evaluation.

        Args:
        - candidates: A Dataset or DatasetDict containing the candidate passages to score.
        - ground_truth: A Dataset or DatasetDict containing the ground-truth passages.
        - kb: The knowledge base to use for context enrichment.
        - context_length: The maximum character length of the left / right context to use for scoring (default 128 chars).
        - expand_abbreviations: Whether to expand abbreviations in the passages.
        - encode_sem_type: Whether to include the semantic type of the concept in its representation
        - masking: Whether to mask entities in the passages.
        - use_nil: Whether to have an option for NIL in each batch for unlinkable entities

        Returns:
        - IndexedDataset or IndexedDatasetDict containing the encoded passages.
        """
        print("Context length:", context_length)
        print("Use NIL values:", use_nil)

        if type(candidates) == DatasetDict:
            assert type(ground_truth) == DatasetDict
            res = IndexedDatasetDict()
            for split, cand in candidates.items():
                gt = ground_truth[split] if ground_truth else None
                ds, doc_index = create_cross_enc_dataset(
                    cand, gt, kb, context_length, expand_abbreviations, encode_sem_type, masking, use_nil
                )
                res[split] = IndexedDataset(ds, doc_index)
            return res
        else:
            ds, doc_index = create_cross_enc_dataset(
                candidates, ground_truth, kb, context_length, expand_abbreviations, encode_sem_type, masking, use_nil
            )
            return IndexedDataset(ds, doc_index)

    def fit(
        self,
        training_args: CrossEncoderTrainingArgs,
        train_dataset,
        val_dataset,
        output_dir: Union[str, Path] = "./output/cross_encoder",
        train_continue=False,
        loss_fct=None,
        add_special_tokens=True,
        max_length=512,
        eval_callback=None,
        show_progress_bar=True,
    ):
        """
        Fits the model using the given training and validation datasets.

        Args:
        - training_args: Training Arguments
        - train_dataset (List[InputExample]): The list of InputExample objects representing the training dataset.
        - val_dataset (List[InputExample]): The list of InputExample objects representing the validation dataset.
        - output_dir (Union[str, Path]): The directory where the trained model will be saved.
        - train_continue (bool, optional): If True, the training will be continued from the current state of the model. Defaults to False.
        - softmax_loss (bool, optional): If True, uses CrossEntropyLoss as the loss function. Otherwise, no loss function is used. Defaults to True.
        - loss_fct (optional): The loss function to be used. If None, the function will be automatically set based on the value of softmax_loss. Defaults to None.
        - add_special_tokens (bool, optional): If True, additional special tokens are added to the tokenizer. Defaults to True.
        - max_length (int, optional): The maximum length of the input sequence. Defaults to 512.
        - fp16 (bool, optional): If True, uses mixed-precision training. Defaults to False.
        - eval_callback (optional): A callback function to be called during evaluation. Defaults to None.
        - **training_args: A dictionary containing the training arguments to be passed to the ScoredCrossEncoder model.

        Raises:
        - AssertionError: If train_continue is False and the model already exists.
        """
        for k, v in training_args.args.items():
            print(k, ":=", v)
        set_seed(training_args.get("random_seed", 0))
        if not self.model:
            self.model = ScoredCrossEncoder(training_args["model_name"], num_labels=1, max_length=max_length)
            if add_special_tokens:
                self.model.tokenizer.add_special_tokens(
                    {"additional_special_tokens": ["[START]", "[END]", "[TITLE]", "[NIL]", "[TYPE]"]}
                )
                self.model.model.resize_token_embeddings(len(self.model.tokenizer))
        else:
            assert train_continue, "Training must be continued if model is set"
            print("Continue training")

        if not loss_fct:
            if training_args["softmax_loss"]:
                loss_fct = nn.CrossEntropyLoss()
            else:
                loss_fct = None

        # We wrap train_samples (which is a List[InputExample]) into a pytorch DataLoader
        train_dataloader = torch.utils.data.DataLoader(
            BatchedCrossEncoderDataset(train_dataset), num_workers=0, batch_size=None, shuffle=True
        )
        evaluator = EntityLinkingEvaluator(val_dataset, name="eval", eval_callback=eval_callback)

        #### Just some code to print debug information to stdout
        logging.basicConfig(
            format="%(asctime)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            level=logging.INFO,
            handlers=[LoggingHandler()],
        )
        #### /print debug information to stdout

        # Train the model
        self.model.fit_with_scores(
            train_dataloader=train_dataloader,
            evaluator=evaluator,
            epochs=training_args["num_train_epochs"],
            loss_fct=loss_fct,
            warmup_steps=100,
            output_path=output_dir,
            callback=None,
            optimizer_params={"lr": training_args["learning_rate"]},
            use_amp=training_args["fp16"],
            label_smoothing=training_args["label_smoothing"],
            rank_regularization=training_args["rank_regularization"],
            train_layers=training_args["train_layers"],
            show_progress_bar=show_progress_bar,
        )

    def rerank_batch(
        self, candidates, cross_enc_dataset, show_progress_bar=True, convert_to_numpy=True, allow_nil=True
    ):
        """
        Re-ranks a batch of candidates using a cross encoder and returns the re-ranked candidates.

        Args:
        - candidates: a dataset of candidate inputs to be re-ranked.
        - cross_enc_dataset: a dataset of cross-encoder inputs for scoring the candidates.
        - show_progress_bar: a boolean indicating whether to display a progress bar during prediction.
        - convert_to_numpy: a boolean indicating whether to convert predictions to numpy arrays.
        - allow_nil: allow to produce an empty result, if NIL is the most likely option

        Returns:
        - A dataset of re-ranked candidates.
        """
        predictions = _cross_encoder_predict(self.model, cross_enc_dataset.dataset, show_progress_bar, convert_to_numpy)
        reranked = candidates.map(
            lambda d, i: rerank(d, i, cross_enc_dataset.index, cross_enc_dataset.dataset, predictions, allow_nil),
            with_indices=True,
            load_from_cache_file=False,
        )
        return Deduplicator().transform_batch(reranked)


class EntityLinkingEvaluator:
    """
    Evaluates a model on an entity linking dataset and returns the accuracy.

    Args:
    - el_dataset (Dataset): The entity linking dataset to evaluate on.
    - name (str): The name of the dataset being evaluated. Defaults to "-".
    - show_progress_bar (bool): Whether to show a progress bar during evaluation. Defaults to False.
    - eval_callback (function): A callback function to call during evaluation. Defaults to None.
    - k_max (int): The maximum value of k to calculate accuracy at. Defaults to 64.

    Returns:
    - float: The accuracy of the model on the entity linking dataset.
    """

    def __init__(self, el_dataset, name="-", show_progress_bar: bool = False, eval_callback=None, k_max: int = 64):
        self.name = name
        self.el_dataset = el_dataset
        self.show_progress_bar = show_progress_bar
        self.eval_callback = eval_callback
        self.k_max = k_max

    def __call__(self, model, output_path: str = None, epoch: int = -1, steps: int = -1) -> float:
        if epoch != -1:
            if steps == -1:
                out_txt = " after epoch {}:".format(epoch)
            else:
                out_txt = " in epoch {} after {} steps:".format(epoch, steps)
        else:
            out_txt = ":"

        logger.info("EntityLinkingEvaluator: Evaluating the model on " + self.name + " dataset" + out_txt)
        model_pred_scores = _cross_encoder_predict(
            model,
            self.el_dataset,
            show_progress_bar=self.show_progress_bar,
            convert_to_numpy=True,
        )

        def get_accuracy(pred_scores, k=1):
            tp = fp = 0

            for scores, batch in zip(pred_scores, self.el_dataset):
                labels = np.array([b.label for b in batch])
                pred_indices = scores.argsort(kind="stable")[-1::-1]
                top_k = pred_indices[0:k]
                ground_truth = labels.argmax() if (labels > 0).any() else -1

                if any(top_k == ground_truth):
                    tp += 1
                else:
                    fp += 1
            return tp / (tp + fp)

        acc = get_accuracy(model_pred_scores, 1)
        acc5 = get_accuracy(model_pred_scores, 5)
        k_max = self.k_max
        accmax = get_accuracy(model_pred_scores, k_max)

        logger.info(f"Accuracy: {acc}")
        logger.info(f"Accuracy @ 5: {acc5}")
        logger.info(f"Accuracy @ {k_max}: {accmax}")

        if self.eval_callback:
            self.eval_callback(
                {
                    "train/epoch": epoch,
                    f"{self.name}/accuracy_1": acc,
                    f"{self.name}/accuracy_5": acc5,
                    f"{self.name}/accuracy_{k_max}": accmax,
                }
            )

        baseline_pred_scores = np.array([[0]] * len(self.el_dataset))
        baseline_acc = get_accuracy(baseline_pred_scores, 1)
        logger.info(f"Baseline Accuracy: {baseline_acc}")

        return acc


class BatchedCrossEncoderDataset(torch.utils.data.Dataset):
    """
    A PyTorch Dataset class for batched cross-encoder data.

    Args:
    - cross_enc_data (list): The batched cross-encoder data.
    """

    def __init__(self, cross_enc_data):
        self.cross_enc_data = cross_enc_data

    def __len__(self):
        return len(self.cross_enc_data)

    def __getitem__(self, idx):
        return self.cross_enc_data[idx]



from transformers import AutoModelForSequenceClassification, AutoTokenizer, AutoConfig
import numpy as np
import logging
import os
from typing import Dict, Type, Callable, List, Union
import transformers
from sentence_transformers.cross_encoder import CrossEncoder
from sentence_transformers.evaluation import SentenceEvaluator
from sentence_transformers import SentenceTransformer

import torch
from torch.utils.data import DataLoader
from torch.optim import Optimizer
from tqdm.autonotebook import tqdm, trange

import re

logger = logging.getLogger(__name__)


class ScoredInputExample:
    """
    Structure for one input example with texts, the label and a unique id
    """

    def __init__(
        self, guid: str = "", texts: List[str] = None, label: Union[int, float] = 0, score: float = None, nil=False
    ):
        """
        Creates one InputExample with the given texts, guid and label
        :param guid
            id for the example
        :param texts
            the texts for the example.
        :param label
            the label for the example
        :score score
            prior score for the example
        :nil nil
            whether the example is a nil example
        """
        self.guid = guid
        self.texts = texts
        self.label = label
        self.score = score
        self.nil = nil

    def __repr__(self):
        return "<ScoredInputExample> label: {}, score: {}, texts: {}, nil: {}".format(
            str(self.label), str(self.score), "; ".join(self.texts), self.nil
        )


class ScoredCrossEncoder(CrossEncoder):
    def __init__(
        self,
        model_name: str,
        num_labels: int = None,
        max_length: int = None,
        device: str = None,
        tokenizer_args: Dict = {},
        automodel_args: Dict = {},
        default_activation_function=None,
    ):
        super().__init__(
            model_name, num_labels, max_length, device, tokenizer_args, automodel_args, default_activation_function
        )

    def smart_batching_collate_with_scores(self, batch):
        texts = [[] for _ in range(len(batch[0].texts))]
        labels = []
        scores = []

        for example in batch:
            for idx, text in enumerate(example.texts):
                texts[idx].append(text.strip())
            labels.append(example.label)
            scores.append(example.score)

        tokenized = self.tokenizer(
            *texts, padding=True, truncation="longest_first", return_tensors="pt", max_length=self.max_length
        )
        labels = torch.tensor(labels, dtype=torch.float if self.config.num_labels == 1 else torch.long).to(
            self._target_device
        )
        scores = torch.tensor(scores, dtype=torch.float).to(self._target_device)

        for name in tokenized:
            tokenized[name] = tokenized[name].to(self._target_device)

        return tokenized, labels, scores

    def rank_regularizer(self, logits, scores):
        pdist = torch.nn.PairwiseDistance(p=2)
        return pdist(logits, scores)

    def fit_with_scores(
        self,
        train_dataloader: DataLoader,
        evaluator: SentenceEvaluator = None,
        epochs: int = 1,
        loss_fct=None,
        activation_fct=torch.nn.Identity(),
        scheduler: str = "WarmupLinear",
        warmup_steps: int = 10000,
        optimizer_class: Type[Optimizer] = torch.optim.AdamW,
        optimizer_params: Dict[str, object] = {"lr": 2e-5},
        weight_decay: float = 0.01,
        rank_regularization: bool = 1.0,
        train_layers: list = [],
        label_smoothing: float = 0.0,
        evaluation_steps: int = 0,
        output_path: str = None,
        save_best_model: bool = True,
        max_grad_norm: float = 1,
        use_amp: bool = False,
        callback: Callable[[Dict], None] = None,
        show_progress_bar: bool = True,
    ):
        """
        Train the model with the given training objective
        Each training objective is sampled in turn for one batch.
        We sample only as many batches from each objective as there are in the smallest one
        to make sure of equal training with each dataset.

        :param train_dataloader: DataLoader with training InputExamples
        :param evaluator: An evaluator (sentence_transformers.evaluation) evaluates the model performance during training on held-out dev data. It is used to determine the best model that is saved to disc.
        :param epochs: Number of epochs for training
        :param loss_fct: Which loss function to use for training. If None, will use nn.BCEWithLogitsLoss() if self.config.num_labels == 1 else nn.CrossEntropyLoss()
        :param activation_fct: Activation function applied on top of logits output of model.
        :param scheduler: Learning rate scheduler. Available schedulers: constantlr, warmupconstant, warmuplinear, warmupcosine, warmupcosinewithhardrestarts
        :param warmup_steps: Behavior depends on the scheduler. For WarmupLinear (default), the learning rate is increased from o up to the maximal learning rate. After these many training steps, the learning rate is decreased linearly back to zero.
        :param optimizer_class: Optimizer
        :param optimizer_params: Optimizer parameters
        :param weight_decay: Weight decay for model parameters
        :param evaluation_steps: If > 0, evaluate the model using evaluator after each number of training steps
        :param output_path: Storage path for the model and evaluation files
        :param save_best_model: If true, the best model (according to evaluator) is stored at output_path
        :param max_grad_norm: Used for gradient normalization.
        :param use_amp: Use Automatic Mixed Precision (AMP). Only for Pytorch >= 1.6.0
        :param callback: Callback function that is invoked after each evaluation.
        :param show_progress_bar: If True, output a tqdm progress bar
        """
        train_dataloader.collate_fn = self.smart_batching_collate_with_scores

        if use_amp:
            from torch.cuda.amp import autocast

            scaler = torch.cuda.amp.GradScaler()

        self.model.to(self._target_device)

        if output_path is not None:
            os.makedirs(output_path, exist_ok=True)

        self.best_score = -9999999
        num_train_steps = int(len(train_dataloader) * epochs)

        # Prepare optimizers
        param_optimizer = list(self.model.named_parameters())

        no_decay = ["bias", "LayerNorm.bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in param_optimizer if not any(nd in n for nd in no_decay)],
                "weight_decay": weight_decay,
            },
            {"params": [p for n, p in param_optimizer if any(nd in n for nd in no_decay)], "weight_decay": 0.0},
        ]

        if train_layers:
            print("Training only layers: ", train_layers)
            for n, param in param_optimizer:
                if not any([n.startswith(f) or re.match(f, n) for f in train_layers]):
                    print("Freezing", n)
                    param.requires_grad = False

        optimizer = optimizer_class(optimizer_grouped_parameters, **optimizer_params)

        if isinstance(scheduler, str):
            scheduler = SentenceTransformer._get_scheduler(
                optimizer, scheduler=scheduler, warmup_steps=warmup_steps, t_total=num_train_steps
            )

        if loss_fct is None:
            loss_fct = torch.nn.BCEWithLogitsLoss() if self.config.num_labels == 1 else torch.nn.CrossEntropyLoss()

        skip_scheduler = False
        for epoch in trange(epochs, desc="Epoch", disable=not show_progress_bar):
            training_steps = 0
            self.model.zero_grad()
            self.model.train()

            for features, labels, scores in tqdm(
                train_dataloader, desc="Iteration", smoothing=0.05, disable=not show_progress_bar
            ):
                if label_smoothing > 0:
                    labels = (1 - label_smoothing) * labels + (label_smoothing / len(labels))
                if use_amp:
                    with autocast():
                        model_predictions = self.model(**features, return_dict=True)
                        logits = activation_fct(model_predictions.logits)
                        if self.config.num_labels == 1:
                            logits = logits.view(-1)
                        loss_value = loss_fct(logits, labels)
                        if rank_regularization != 0:
                            loss_value += rank_regularization * self.rank_regularizer(logits, scores)

                    scale_before_step = scaler.get_scale()
                    scaler.scale(loss_value).backward()
                    scaler.unscale_(optimizer)
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                    scaler.step(optimizer)
                    scaler.update()

                    skip_scheduler = scaler.get_scale() != scale_before_step
                else:
                    model_predictions = self.model(**features, return_dict=True)
                    logits = activation_fct(model_predictions.logits)
                    if self.config.num_labels == 1:
                        logits = logits.view(-1)
                    loss_value = loss_fct(logits, labels)
                    if rank_regularization != 0:
                        loss_value += rank_regularization * self.rank_regularizer(logits, scores)
                    loss_value.backward()
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                    optimizer.step()

                optimizer.zero_grad()

                if not skip_scheduler:
                    scheduler.step()

                training_steps += 1

                if evaluator is not None and evaluation_steps > 0 and training_steps % evaluation_steps == 0:
                    self._eval_during_training(evaluator, output_path, save_best_model, epoch, training_steps, callback)

                    self.model.zero_grad()
                    self.model.train()

            if evaluator is not None:
                self._eval_during_training(evaluator, output_path, save_best_model, epoch, -1, callback)

    def predict(
        self,
        sentences: List[List[str]],
        batch_size: int = 32,
        show_progress_bar: bool = None,
        num_workers: int = 0,
        activation_fct=None,
        apply_softmax=False,
        convert_to_numpy: bool = True,
        convert_to_tensor: bool = False,
    ):
        """
        Performs predicts with the CrossEncoder on the given sentence pairs.

        :param sentences: A list of sentence pairs [[Sent1, Sent2], [Sent3, Sent4]]
        :param batch_size: Batch size for encoding
        :param show_progress_bar: Output progress bar
        :param num_workers: Number of workers for tokenization
        :param activation_fct: Activation function applied on the logits output of the CrossEncoder. If None, nn.Sigmoid() will be used if num_labels=1, else nn.Identity
        :param convert_to_numpy: Convert the output to a numpy matrix.
        :param apply_softmax: If there are more than 2 dimensions and apply_softmax=True, applies softmax on the logits output
        :param convert_to_tensor:  Conver the output to a tensor.
        :return: Predictions for the passed sentence pairs
        """
        input_was_string = False
        if isinstance(sentences[0], str):  # Cast an individual sentence to a list with length 1
            sentences = [sentences]
            input_was_string = True

        inp_dataloader = DataLoader(
            sentences,
            batch_size=batch_size,
            collate_fn=self.smart_batching_collate_text_only,
            num_workers=num_workers,
            shuffle=False,
        )

        if show_progress_bar is None:
            show_progress_bar = (
                logger.getEffectiveLevel() == logging.INFO or logger.getEffectiveLevel() == logging.DEBUG
            )

        iterator = inp_dataloader
        if show_progress_bar:
            iterator = tqdm(inp_dataloader, desc="Batches")

        if activation_fct is None:
            activation_fct = self.default_activation_function

        pred_scores = []
        self.model.eval()
        self.model.to(self._target_device)
        with torch.no_grad():
            for features in iterator:
                model_predictions = self.model(**features, return_dict=True)
                logits = activation_fct(model_predictions.logits)

                if apply_softmax and len(logits[0]) > 1:
                    logits = torch.nn.functional.softmax(logits, dim=1)
                pred_scores.extend(logits)

        if self.config.num_labels == 1:
            pred_scores = [score[0] for score in pred_scores]

        if convert_to_tensor:
            pred_scores = torch.stack(pred_scores)
        elif convert_to_numpy:
            pred_scores = np.asarray([score.cpu().detach().numpy() for score in pred_scores])

        if input_was_string:
            pred_scores = pred_scores[0]

        return pred_scores

    def _eval_during_training(self, evaluator, output_path, save_best_model, epoch, steps, callback):
        """Runs evaluation during the training"""
        if evaluator is not None:
            score = evaluator(self, output_path=output_path, epoch=epoch, steps=steps)
            if callback is not None:
                callback(score, epoch, steps)
            if score > self.best_score:
                self.best_score = score
                if save_best_model:
                    self.save(output_path)

    def save(self, path):
        """
        Saves all model and tokenizer to path
        """
        if path is None:
            return

        logger.info("Save model to {}".format(path))
        self.model.save_pretrained(path)
        self.tokenizer.save_pretrained(path)

    def save_pretrained(self, path):
        """
        Same function as save
        """
        return self.save(path)



from xmen.ext.scispacy.umls_utils import (
    read_umls_file_headers,
    read_umls_concepts,
    read_umls_types,
    read_umls_definitions,
)
from scispacy import umls_semantic_type_tree
from scispacy.linking_utils import DEFAULT_UMLS_TYPES_PATH
from langcodes import Language
import pandas as pd
from pathlib import Path
import os
from .log import logger


def read_umls_sabs(meta_path):
    """
    Reads the Metathesaurus source abbreviation (SAB) information from the UMLS data file "MRSAB.RRF"
    located at `meta_path` directory and returns it as a pandas DataFrame.

    Args:
    - meta_path (str): Path to the directory containing the UMLS data files.

    Returns:
    - pandas.DataFrame: A DataFrame with the following columns: 'RSAB', 'VSAB', 'SON', 'SF', 'TTY', 'ATN',
    'CVF', 'TS', 'LUI', 'STY', 'SRL', 'SUPPRESS', 'CVF_CUI'.

    Raises:
    - AssertionError: If the number of headers in the data file is not equal to the number of splits in a line.
    """
    res = []
    sab_filename = "MRSAB.RRF"
    headers = read_umls_file_headers(meta_path, sab_filename)
    with open(f"{meta_path}/{sab_filename}") as fin:
        for line in fin:
            splits = line.strip().split("|")
            assert len(headers) == len(splits)
            sabs = dict(zip(headers, splits))
            res.append(sabs)
    return pd.DataFrame(res)


def get_semantic_groups(version: str = None):
    """
    Retrieves the semantic groups and their corresponding TUIs (Type Unique Identifiers) from the
    UMLS semantic groups file "SemGroups.txt" located in the `resources` directory.

    Args:
    - version (str, optional): Version of the UMLS semantic groups to retrieve. Defaults to None.

    Returns:
    - pandas.DataFrame: A DataFrame with the following columns: 'GRP', 'GRP_NAME', 'TUI', 'TUI_NAME'.
    """
    sem_group_path = (
        Path(os.path.dirname(os.path.abspath(__file__)))
        / "resources"
        / f'SemGroups{"-" + version if version else ""}.txt'
    )
    sem_groups = pd.read_csv(sem_group_path, sep="|")
    sem_groups.columns = ["GRP", "GRP_NAME", "TUI", "TUI_NAME"]
    return sem_groups


def filter_semantic_groups(semantic_groups, concept_details, version=None):
    """
    Filters the `concept_details` dictionary by keeping only the concepts whose TUIs (Type Unique Identifiers)
    belong to the specified semantic groups.

    Args:
    - semantic_groups (list of str): List of semantic group names to filter by.
    - concept_details (dict): Dictionary with concept IDs as keys and their details as values.
    - version (str, optional): Version of the UMLS semantic groups to use for filtering. Defaults to None.

    Returns:
    - dict: A filtered version of the `concept_details` dictionary containing only concepts whose TUIs
    belong to the specified semantic groups.
    """
    sem_groups = get_semantic_groups(version)
    logger.info(f"> Filtering by semantic groups with version {version} and {len(sem_groups)} types.")
    valid_tuis = sem_groups[sem_groups.GRP.isin(semantic_groups)].TUI.unique()
    return {k: v for k, v in concept_details.items() if any([t in valid_tuis for t in v["types"]])}


def get_sem_type_tree():
    """
    Constructs a UMLS semantic type tree from the TSV file located at the default UMLS types path
    and returns it as a UmlsSemanticTypeTree object.

    Returns:
    - UmlsSemanticTypeTree: A UmlsSemanticTypeTree object representing the UMLS semantic type tree.
    """
    return umls_semantic_type_tree.construct_umls_tree_from_tsv(DEFAULT_UMLS_TYPES_PATH)


def filter_semantic_types(tuis, expand_semantic_types, concept_details):
    """
    Filters the `concept_details` dictionary by keeping only the concepts whose TUIs (Type Unique Identifiers)
    belong to the specified semantic types.

    Args:
    - tuis (list of str): List of semantic type abbreviations to filter by.
    - expand_semantic_types (bool): Whether or not to include child semantic types of the specified ones.
    - concept_details (dict): Dictionary with concept IDs as keys and their details as values.

    Returns:
    - dict: A filtered version of the `concept_details` dictionary containing only concepts whose TUIs
    belong to the specified semantic types.
    """
    if expand_semantic_types:
        tuis = expand_tuis(tuis, get_sem_type_tree())
    return {k: v for k, v in concept_details.items() if any([t in tuis for t in v["types"]])}


def get_alias_count(concept_details):
    """
    Calculates and returns the total number of aliases for all the concepts in the `concept_details` dictionary.

    Args:
    - concept_details (dict): Dictionary with concept IDs as keys and their details as values.

    Returns:
    - int: Total number of aliases for all the concepts in the `concept_details` dictionary.
    """
    return sum([len(c["aliases"]) + 1 for c in concept_details.values()])


def expand_tuis(tuis, sem_type_tree):
    """
    Recursively expands a list of UMLS semantic type abbreviations to include their child semantic types,
    using the specified semantic type tree.

    Args:
    - tuis (list of str): List of semantic type abbreviations to expand.
    - sem_type_tree (UMLSSemanticTypeTree): A UMLSSemanticTypeTree object representing the UMLS semantic type tree.

    Returns:
    - list of str: A list of semantic type abbreviations that includes the specified types and all their child types.
    """
    result = tuis
    for t in tuis:
        children = [c.type_id for c in sem_type_tree.type_id_to_node[t].children]
        if len(children) > 0:
            result += children
            result += expand_tuis(children, sem_type_tree)
    return list(set(result))


def _get_lang_code(lang):
    """
    Converts a language name or ISO code to its ISO 639-2/B code.

    Args:
    - lang (str): Language name or ISO code.

    Returns:
    - str: ISO 639-2/B code for the specified language.

    Raises:
    - AssertionError: If the specified language is not valid.
    """
    lang_obj = Language.get(lang)
    assert lang_obj.is_valid()
    return lang_obj.to_alpha3(variant="B").upper()


def get_umls_concepts(
    meta_path: str,
    langs: list,
    sabs: list,
    sources: list,
    semantic_groups: list,
    semantic_types: list,
    expand_semantic_types=True,
    non_suppressed_only=False,
    semantic_group_file_version=None,
) -> dict:
    """
    Reads UMLS concepts and related metadata from the specified sources and languages, and filters them by semantic
    groups and semantic types.

    Args:
    - meta_path (str): Path to the UMLS metathesaurus directory.
    - langs (list): A list of language codes to include. If `None` or `'all'`, all available languages
                   will be included.
    - sabs (list): A list of source abbreviations to include. If empty, all available sources will be included.
    - sources (list): A list of source full names to include. If empty, all available sources will be included.
    - semantic_groups (list): A list of semantic group abbreviations to include. If empty, no semantic group filtering
                              will be performed.
    - semantic_types (list): A list of semantic type abbreviations to include. If empty, no semantic type filtering
                             will be performed.
    - expand_semantic_types (bool): Whether to expand the specified semantic types to include all child types.
    - non_suppressed_only (bool): Whether to include only non-suppressed concepts.
    - semantic_group_file_version (str): The version of the UMLS Semantic Groups file to use. If `None`, the latest
                                         version will be used.

    Returns:
    - dict: A dictionary containing the UMLS concepts and related metadata, filtered according to the specified
            criteria.
    """

    meta_path = Path(meta_path)
    if not langs or "all" in langs:
        langs = None
    else:
        langs = [_get_lang_code(lang) for lang in langs]

    logger.info(f"Using UMLS metathesaurus: {meta_path}")

    if not sabs:
        sabs = []

    if sources:
        sab_df = read_umls_sabs(meta_path)
        sabs += list(sab_df[sab_df.SF.isin(sources)].RSAB.unique())

    logger.info(f"Using sources: {sabs}")

    concept_details = {}
    if not sabs:
        sabs = [None]
    for source in sabs:
        logger.info(
            f'>> Reading concepts from {"all sources" if not source else source} and {"all languages" if not langs else f"languages: {langs}"}'
        )
        for lang in langs if langs else [None]:
            read_umls_concepts(meta_path, concept_details, source=source, lang=lang, non_suppressed=non_suppressed_only)

    logger.info(">> Reading types ... ")
    read_umls_types(meta_path, concept_details)

    if semantic_groups:
        logger.info(f"> Number of concepts before semantic group filtering: {len(concept_details)}")
        concept_details = filter_semantic_groups(semantic_groups, concept_details, semantic_group_file_version)

    if semantic_types:
        logger.info(f"> Number of concepts before semantic type filtering: {len(concept_details)}")
        concept_details = filter_semantic_types(semantic_types, expand_semantic_types, concept_details)

    logger.info(">> Reading definitions ... ")
    read_umls_definitions(meta_path, concept_details)

    logger.info(f"> Number of concepts before de-duplication: {len(concept_details)}")
    logger.info(f"> Number of aliases before de-duplication: {get_alias_count(concept_details)}")

    for concept in concept_details.values():
        # Some concepts have many duplicate aliases. Here we remove them.
        concept["aliases"] = list(set(concept["aliases"]))

        # if a concept doesn't have a canonical name, use the first alias instead
        if "canonical_name" not in concept:
            aliases = concept["aliases"]
            concept["canonical_name"] = aliases[0]
            del aliases[0]

        # deleting `is_from_preferred_source`
        if "is_from_preferred_source" in concept:
            del concept["is_from_preferred_source"]

    return concept_details



"""
This script is a modified version of a script originally licensed under the Apache 2.0 license.

Original script: https://github.com/allenai/scispacy/blob/main/scispacy/umls_utils.py
Original authors: Neumann, Mark and King, Daniel and Beltagy, Iz and Ammar, Waleed

Modifications: added two new arguments `lang` and `non_suppressed` to `read_umls_concepts` 
(with their respective default values) which allow the function to filter the terms by language 
and decide whether to keep only non-suppressed concepts or not.
"""

from typing import Optional, List, Dict

# TODO(Mark): Remove in scispacy v1.0, for backward compatability only.
from scispacy.linking_utils import Entity as UmlsEntity, UmlsKnowledgeBase  # noqa

# preferred definition sources (from S2)
DEF_SOURCES_PREFERRED = {"NCI_BRIDG", "NCI_NCI-GLOSS", "NCI", "GO", "MSH", "NCI_FDA"}


def read_umls_file_headers(meta_path: str, filename: str) -> List[str]:
    """
    Read the file descriptor MRFILES.RRF from a UMLS release and get column headers (names)
    for the given file

    MRFILES.RRF file format: a pipe-separated values
    Useful columns:
        column 0: name of one of the files in the META directory
        column 2: column names of that file

    Args:
        meta_path: path to the META directory of an UMLS release
        filename: name of the file to get its column headers
    Returns:
        a list of column names
    """
    file_descriptors = f"{meta_path}/MRFILES.RRF"  # to get column names
    with open(file_descriptors, encoding="utf-8") as fin:
        for line in fin:
            splits = line.split("|")
            found_filename = splits[0]
            column_names = (splits[2] + ",").split(
                ","
            )  # ugly hack because all files end with an empty column
            if found_filename in filename:
                return column_names
    assert False, f"Couldn't find column names for file {filename}"
    return None


def read_umls_concepts(
    meta_path: str,
    concept_details: Dict,
    source: Optional[str] = None,
    lang: str = "ENG",
    non_suppressed: bool = True,
):
    """
    Read the concepts file MRCONSO.RRF from a UMLS release and store it in
    concept_details dictionary. Each concept is represented with
    - concept_id
    - canonical_name
    - aliases
    - types
    - definition
    This function fills the first three. If a canonical name is not found, it is left empty.

    MRFILES.RRF file format: a pipe-separated values
    Useful columns: CUI, LAT, SUPPRESS, STR, ISPREF, TS, STT

    Args:
        meta_path: path to the META directory of an UMLS release
        concept_details: a dictionary to be filled with concept informations
        source: An optional source identifier, used as a filter to extract only a
                specific source from UMLS.
        lang: An optional language identifier, used to filter terms by language
        non_suppressed: flag to indicate whether only non-suppressed concepts should be kept
    """
    concepts_filename = "MRCONSO.RRF"
    headers = read_umls_file_headers(meta_path, concepts_filename)
    with open(f"{meta_path}/{concepts_filename}", encoding="utf-8") as fin:
        for line in fin:
            splits = line.strip().split("|")
            assert len(headers) == len(splits), (headers, splits)
            concept = dict(zip(headers, splits))
            if (lang is not None and concept["LAT"] != lang) or (
                non_suppressed and concept["SUPPRESS"] != "N"
            ):
                continue  # Keep non-suppressed concepts in target language only

            if source is not None:
                if concept["SAB"] != source:
                    continue

            concept_id = concept["CUI"]
            if concept_id not in concept_details:  # a new concept
                # add it to the dictionary with an empty list of aliases and types
                concept_details[concept_id] = {
                    "concept_id": concept_id,
                    "aliases": [],
                    "types": [],
                }

            concept_name = concept["STR"]
            # this condition is copied from S2. It checks if the concept name is canonical or not
            is_canonical = (
                concept["ISPREF"] == "Y"
                and concept["TS"] == "P"
                and concept["STT"] == "PF"
            )

            if not is_canonical or "canonical_name" in concept_details[concept_id]:
                # not a canonical name or a canonical name already found
                concept_details[concept_id]["aliases"].append(
                    concept_name
                )  # add it as an alias
            else:
                concept_details[concept_id][
                    "canonical_name"
                ] = concept_name  # set as canonical name


def read_umls_types(meta_path: str, concept_details: Dict):
    """
    Read the types file MRSTY.RRF from a UMLS release and store it in
    concept_details dictionary. This function adds the `types` field
    to the information of each concept

    MRSTY.RRF file format: a pipe-separated values
    Useful columns: CUI, TUI

    Args:
        meta_path: path to the META directory of an UMLS release
        concept_details: a dictionary to be filled with concept informations
    """
    types_filename = "MRSTY.RRF"
    headers = read_umls_file_headers(meta_path, types_filename)
    with open(f"{meta_path}/{types_filename}", encoding="utf-8") as fin:
        for line in fin:
            splits = line.strip().split("|")
            assert len(headers) == len(splits)
            concept_type = dict(zip(headers, splits))

            concept = concept_details.get(concept_type["CUI"])
            if (
                concept is not None
            ):  # a small number of types are for concepts that don't exist
                concept["types"].append(concept_type["TUI"])


def read_umls_definitions(meta_path: str, concept_details: Dict):
    """
    Read the types file MRDEF.RRF from a UMLS release and store it in
    concept_details dictionary. This function adds the `definition` field
    to the information of each concept

    MRDEF.RRF file format: a pipe-separated values
    Useful columns: CUI, SAB, SUPPRESS, DEF

    Args:
        meta_path: path to the META directory of an UMLS release
        concept_details: a dictionary to be filled with concept informations
    """
    definitions_filename = "MRDEF.RRF"
    headers = read_umls_file_headers(meta_path, definitions_filename)
    with open(f"{meta_path}/{definitions_filename}", encoding="utf-8") as fin:
        headers = read_umls_file_headers(meta_path, definitions_filename)
        for line in fin:
            splits = line.strip().split("|")
            assert len(headers) == len(splits)
            definition = dict(zip(headers, splits))

            if definition["SUPPRESS"] != "N":
                continue
            is_from_preferred_source = definition["SAB"] in DEF_SOURCES_PREFERRED
            concept = concept_details.get(definition["CUI"])
            if (
                concept is None
            ):  # a small number of definitions are for concepts that don't exist
                continue

            if (
                "definition" not in concept
                or is_from_preferred_source
                and concept["is_from_preferred_source"] == "N"
            ):
                concept["definition"] = definition["DEF"]
                concept["is_from_preferred_source"] = (
                    "Y" if is_from_preferred_source else "N"
                )


from omegaconf import DictConfig, OmegaConf
from ..log import logger
from xmen.kb import create_flat_term_dict
from xmen.linkers import TFIDFNGramLinker, SapBERTLinker
from pathlib import Path
import torch

DEFAULT_BATCH_SIZE = 2048
DEFAULT_INDEX_BUFFER_SIZE = 50000


def build_ngram(cfg: DictConfig, work_dir: Path, dict_dir: Path):
    """Build an N-Gram index for the given dictionary directory and store it in the working directory.

    Args:
    - cfg (DictConfig): The configuration for building the index.
    - work_dir (Path): The working directory where the index will be stored.
    - dict_dir (Path): The directory containing the dictionary file.
    """

    # ensure the index folder exists
    work_dir.parent.mkdir(exist_ok=True, parents=True)

    # N-Gram indices
    TFIDFNGramLinker.write_index(work_dir / "index" / "ngrams", [dict_dir])
    pass


def build_sapbert(
    cfg: DictConfig,
    work_dir: Path,
    dict_dir: Path,
    gpu_id: int,
    batch_size: int,
    index_buffer_size: int,
    save_ram: bool,
):
    """Builds an index of concept embeddings using SapBERT.

    Args:
    - cfg (DictConfig): Hydra configuration object.
    - work_dir (Path): Path to the working directory.
    - dict_dir (Path): Path to the directory containing the concept dictionaries.
    - gpu_id (int): ID of the GPU to be used for training. Use -1 for CPU.
    - batch_size (int): GPU batch size
    - index_buffer_size (int): Buffer size for use for writing SapBERT FAISS Index
    """

    # ensure the index folder exists
    work_dir.parent.mkdir(exist_ok=True, parents=True)
    # SapBERT indices
    term_dict = create_flat_term_dict([dict_dir])
    logger.info(f"Number of aliases: {len(term_dict)}.")
    logger.info(f"Number of concepts: {len(term_dict.cui.unique())}.")

    if cfg.get("linker", {}).get("candidate_generation", {}).get("sapbert", {}).get("model_name", {}) == {}:
        model_name = SapBERTLinker.CROSS_LINGUAL  # default model
    else:
        model_name = cfg.linker.candidate_generation.sapbert.model_name

    cuda = False if gpu_id == -1 else True
    with torch.cuda.device(gpu_id):
        SapBERTLinker.write_index(
            work_dir / "index" / "sapbert",
            term_dict=term_dict,
            cuda=cuda,
            subtract_mean=False,
            model_name=model_name,
            batch_size=batch_size,
            index_buffer_size=index_buffer_size,
            write_memory_map=save_ram,
        )



from collections import defaultdict
import orjson
from pathlib import Path
from typing import List, Union
import pandas as pd
from scispacy.linking_utils import KnowledgeBase, Entity
from collections.abc import Mapping


def load_kb(file_path: Union[str, Path]) -> KnowledgeBase:
    return CompositeKnowledgebase([file_path])


class CompositeKnowledgebase(KnowledgeBase):
    """
    Initializes a CompositeKnowledgebase object with the given list of file paths and optional list of mappers.

    Args:
    - file_paths (List[Union[str, Path]]): A list of file paths for JSONL files that contain the knowledge to be added to the CompositeKnowledgebase.
    - mappers (List[Mapping], optional): A list of mappers to transform each entry in the JSONL files before adding them to the CompositeKnowledgebase. Defaults to None.

    Raises:
    - AssertionError: If the length of the mappers list does not match the length of the file_paths list, or if any of the file paths does not have a ".jsonl" extension.

    Attributes:
    - alias_to_cuis (defaultdict): A defaultdict containing the mapping from each entity alias to the set of CUIs it is associated with.
    - cui_to_entity (dict): A dictionary mapping each CUI to its corresponding Entity object in the CompositeKnowledgebase.
    """

    def __init__(
        self,
        file_paths: List[Union[str, Path]],
        mappers: List[Mapping] = None,
    ):
        if not mappers:
            mappers = [lambda x: x] * len(file_paths)
        assert len(mappers) == len(file_paths)
        alias_to_cuis = defaultdict(set)
        self.cui_to_entity = {}

        for file_path, mapper in zip(file_paths, mappers):
            file_path = Path(file_path)
            assert file_path.suffix == ".jsonl"
            raw = [orjson.loads(line) for line in open(file_path)]

            for entry in raw:
                if mapper:
                    entry = mapper(entry)
                if not entry:
                    continue
                if type(entry) != list:
                    entry = [entry]
                for concept in entry:
                    if type(concept["concept_id"]) == int:
                        concept["concept_id"] = str(concept["concept_id"])
                    cui = concept["concept_id"]
                    unique_aliases = set(concept["aliases"])
                    if "canonical_name" in concept:
                        unique_aliases.add(concept["canonical_name"])
                    for alias in unique_aliases:
                        alias_to_cuis[alias].add(cui)
                    if not cui in self.cui_to_entity:
                        self.cui_to_entity[cui] = Entity(**concept)
                    else:
                        self.cui_to_entity[cui] = _merge_entities(Entity(**concept), self.cui_to_entity[cui])

            self.alias_to_cuis = {**alias_to_cuis}


def _merge_entities(e1: Entity, e2: Entity):
    """
    Merges two entities and returns the resulting entity. The two entities must have the same concept ID.

    Args:
    - e1 (Entity): The first entity to merge.
    - e2 (Entity): The second entity to merge.

    Returns:
    - Entity: The resulting entity after merging the two input entities.

    Raises:
    - AssertionError: If the two input entities have different concept IDs.
    """
    assert e1.concept_id == e2.concept_id

    canonical_name = e1.canonical_name
    if not canonical_name:
        canonical_name = e2.canonical_name
    definition = e1.definition
    if not definition:
        definition = e2.definition

    aliases = list(set(e1.aliases).union(set(e2.aliases)))
    types = list(set(e1.types).union(set(e2.types)))

    return Entity(e1.concept_id, canonical_name, aliases, types, definition)


def create_flat_term_dict(concept_names_jsonl: List[Union[str, Path]], mappers: List[Mapping] = None):
    """
    Creates a flattened pandas dataframe of UMLS terms and their metadata from a list of UMLS concept names JSON files.

    Args:
    - concept_names_jsonl (List[Union[str, Path]]): A list of file paths to UMLS concept names JSON files.
    - mappers (List[Mapping], optional): A list of mappers to apply to each JSON entry, one per JSON file.
    If not provided, defaults to None.

    Returns:
    - pd.DataFrame: A pandas dataframe containing UMLS term data. The columns are 'cui' (str), 'term' (str),
    'canonical' (str), and 'tuis' (list of str), representing the UMLS concept unique identifier, the term or alias,
    the canonical name, and the associated semantic types, respectively.

    Raises:
    - AssertionError: If the number of mappers provided is not equal to the number of JSON files provided, or if
    an entry does not have a canonical name.
    """
    term_dict = []
    if not mappers:
        mappers = [lambda x: x] * len(concept_names_jsonl)
    assert len(mappers) == len(concept_names_jsonl)
    for jsonl_file, mapper in zip(concept_names_jsonl, mappers):
        with open(jsonl_file) as f:
            for entry in f:
                entry = orjson.loads(entry)
                if mapper != None:
                    entry = mapper(entry)
                    if not entry:
                        continue
                if type(entry) != list:
                    entry = [entry]
                for e in entry:
                    assert e["canonical_name"]
                    cui = e["concept_id"]
                    tuis = e["types"]
                    term_dict.append(
                        {
                            "cui": str(cui),
                            "term": e["canonical_name"],
                            "canonical": e["canonical_name"],
                            "tuis": tuis,
                        }
                    )
                    for alias in e["aliases"]:
                        term_dict.append(
                            {
                                "cui": str(cui),
                                "term": alias,
                                "canonical": e["canonical_name"],
                                "tuis": tuis,
                            }
                        )
    term_dict = pd.DataFrame(term_dict)
    return term_dict.drop_duplicates(subset=["cui", "term"])



import torch
import numpy as np
from tqdm import tqdm

from transformers import (
    AutoTokenizer,
    AutoModel,
)

# Adapted from https://github.com/cambridgeltl/sapbert/blob/main/src/model_wrapper.py

_EMBED_DIM = 768


class Model_Wrapper(object):
    """
    Wrapper class for BERT encoder
    """

    def __init__(self):
        self.tokenizer = None
        self.encoder = None

    def get_dense_encoder(self):
        assert self.encoder is not None
        return self.encoder

    def get_dense_tokenizer(self):
        assert self.tokenizer is not None

        return self.tokenizer

    def save_model(self, path, context=False):
        # save bert model, bert config
        self.encoder.save_pretrained(path)

        # save bert vocab
        self.tokenizer.save_pretrained(path)

    def load_model(self, path, max_length=25, use_cuda=True, lowercase=True):
        self.load_bert(path, max_length, use_cuda)

        return self

    def load_bert(self, path, max_length, use_cuda, lowercase=True):
        self.tokenizer = AutoTokenizer.from_pretrained(path, use_fast=True, do_lower_case=lowercase)
        self.encoder = AutoModel.from_pretrained(path)
        if use_cuda:
            self.encoder = self.encoder.cuda()

        return self.encoder, self.tokenizer

    def embed_dense(
        self, names, show_progress=False, use_cuda=True, batch_size=2048, agg_mode="cls", memory_map_file=None
    ):
        """
        Embedding data into dense representations

        Parameters
        ----------
        names : np.array
            An array of names

        Returns
        -------
        dense_embeds : list
            A list of dense embeddings
        """
        self.encoder.eval()  # prevent dropout

        shape = (len(names), _EMBED_DIM)
        batch_size = batch_size
        if memory_map_file:
            dense_embeds = np.memmap(memory_map_file, dtype="float32", mode="w+", shape=shape)
        else:
            dense_embeds = np.zeros(shape, dtype=np.float32)

        with torch.no_grad():
            if show_progress:
                iterations = tqdm(range(0, len(names), batch_size))
            else:
                iterations = range(0, len(names), batch_size)

            for start in iterations:
                end = min(start + batch_size, len(names))
                batch = names[start:end]
                batch_tokenized_names = self.tokenizer.batch_encode_plus(
                    batch,
                    add_special_tokens=True,
                    truncation=True,
                    max_length=25,
                    padding="max_length",
                    return_tensors="pt",
                )

                # generate cuda names only if use cuda
                # and overwrite original batch_tokenized_names
                if use_cuda:
                    batch_tokenized_names_cuda = {}
                    for k, v in batch_tokenized_names.items():
                        batch_tokenized_names_cuda[k] = v.cuda()
                    batch_tokenized_names = batch_tokenized_names_cuda

                # batch_tokenized_names is passed (-_cuda has been removed)
                last_hidden_state = self.encoder(**batch_tokenized_names)[0]
                if agg_mode == "cls":
                    batch_dense_embeds = last_hidden_state[:, 0, :]  # [CLS]
                elif agg_mode == "mean_all_tok":
                    batch_dense_embeds = last_hidden_state.mean(1)  # pooling
                elif agg_mode == "mean":
                    batch_dense_embeds = (
                        last_hidden_state * batch_tokenized_names["attention_mask"].unsqueeze(-1)
                    ).sum(1) / batch_tokenized_names["attention_mask"].sum(-1).unsqueeze(-1)
                else:
                    print("no such agg_mode:", agg_mode)

                batch_dense_embeds = batch_dense_embeds.cpu().detach().numpy()
                dense_embeds[start:end, :] = batch_dense_embeds
        if memory_map_file:
            dense_embeds.flush()
            dense_embeds = np.memmap(memory_map_file, dtype="float32", mode="r", shape=shape)

        return dense_embeds



import pickle
from pathlib import Path
import datasets


class IndexedDatasetDict(dict):
    """
    A dictionary of IndexedDataset objects that can be saved to and loaded from disk.

    Inherits from dict.

    Methods:
    - save_to_disk(folder: str) -> None: Saves the IndexedDataset objects to the specified folder.
    - load_from_disk(folder: str) -> IndexedDatasetDict: Loads the IndexedDataset objects from the specified folder.

    """

    def save_to_disk(self, folder):
        for k, v in self.items():
            v.save_to_disk(Path(folder) / k)

    @staticmethod
    def load_from_disk(folder):
        res = IndexedDatasetDict()
        for f in Path(folder).glob("*"):
            if f.is_dir():
                res[f.name] = IndexedDataset.load_from_disk(f)
        return res


class IndexedDataset:
    """
    A class that wraps a dataset and an index and provides methods to save and load them from disk.

    Attributes:
    - dataset: The dataset to wrap.
    - index: The index of the dataset.

    Methods:
    - save_to_disk(folder: str) -> None: Saves the dataset and the index to the specified folder.
    - load_from_disk(folder: str) -> IndexedDataset: Loads the dataset and the index from the specified folder.
    - __repr__() -> str: Returns a string representation of the IndexedDataset object.
    """

    def __init__(self, dataset, index):
        self.dataset = dataset
        self.index = index

    def save_to_disk(self, folder):
        """
        Saves the dataset and the index to the specified folder.

        Args:
        - folder (str or Path): The path to the folder where the dataset and index will be saved.
        """
        folder = Path(folder)
        folder.mkdir(exist_ok=True, parents=True)
        with open(folder / "dataset", "wb") as fh:
            pickle.dump(self.dataset, fh)
        with open(folder / "index", "wb") as fh:
            pickle.dump(self.index, fh)

    @staticmethod
    def load_from_disk(folder):
        """
        Loads the dataset and the index from the specified folder.

        Args:
        - folder (str or Path): The path to the folder where the dataset and index will be saved.
        """
        with open(Path(folder) / "dataset", "rb") as fh:
            dataset = pickle.load(fh)
        with open(Path(folder) / "index", "rb") as fh:
            index = pickle.load(fh)
        return IndexedDataset(dataset, index)

    def __repr__(self):
        if type(self.dataset) == datasets.Dataset:
            return self.dataset.__repr__()
        else:
            return f"[{len(self.dataset)} items]"



import re
import os
import glob
import numpy as np
import random
import random
import pandas as pd
import json
import torch
from torch.utils.data import Dataset
import logging
from tqdm import tqdm
LOGGER = logging.getLogger(__name__)


class QueryDataset_COMETA(Dataset):

    def __init__(self, data_dir, 
                load_full_sentence=False,
                filter_duplicate=False
        ):
        """       
        Parameters
        ----------
        data_dir : str
            a path of data
        filter_duplicate : bool
            filter duplicate queries
        draft : bool
            use subset of queries for debugging (default False)     
        """
        LOGGER.info("QueryDataset! data_dir={} filter_composite={} filter_duplicate={}".format(
            data_dir, load_full_sentence, filter_duplicate
        ))
        
        self.data = self.load_data(
            data_dir=data_dir,
            load_full_sentence=load_full_sentence,
            filter_duplicate=filter_duplicate
        )
        
    def load_data(self, data_dir, load_full_sentence, filter_duplicate):
        """       
        Parameters
        ----------
        data_dir : str
            a path of data
        filter_composite : bool
            filter composite mentions
        filter_duplicate : bool
            filter duplicate queries  
        
        Returns
        -------
        data : np.array 
            mention, cui pairs
        """
        data = []

        #concept_files = glob.glob(os.path.join(data_dir, "*.concept"))
        data_table = pd.read_csv(data_dir, sep='\t', encoding='utf8')

        for row in data_table.iterrows():
            mention = row[1]["Term"]
            sentence = row[1]["Example"]
            
            #print (mention)
            #print (sentence)

            cui = row[1]["General SNOMED ID"] # TODO: allow general/specific options
            if load_full_sentence: 
                data.append((mention, sentence, cui))
            else:
                data.append((mention, cui))

        if filter_duplicate:
            data = list(dict.fromkeys(data))
        print ("query size:",len(data))
        
        # return np.array data
        data = np.array(data)
        
        return data

class QueryDataset_custom(Dataset):

    def __init__(self, data_dir, 
                load_full_sentence=False,
                filter_duplicate=False
        ):
        """       
        Parameters
        ----------
        data_dir : str
            a path of data
        filter_duplicate : bool
            filter duplicate queries
        draft : bool
            use subset of queries for debugging (default False)     
        """
        LOGGER.info("QueryDataset! data_dir={} filter_duplicate={}".format(
            data_dir, filter_duplicate
        ))
        
        self.data = self.load_data(
            data_dir=data_dir,
            filter_duplicate=filter_duplicate
        )
        
    def load_data(self, data_dir, filter_duplicate):
        """       
        Parameters
        ----------
        data_dir : str
            a path of data
        filter_composite : bool
            filter composite mentions
        filter_duplicate : bool
            filter duplicate queries  
        
        Returns
        -------
        data : np.array 
            mention, cui pairs
        """
        data = []

        with open(data_dir, "r") as f:
            lines = f.readlines()

        for line in lines:
            line = line.rstrip("\n")
            if len(line.split("||")) == 2:
                _id, mention = line.split("||")
            elif len(line.split("||")) == 3: # in case using data with contexts
                _id, mention, context = line.split("||")
            else:
                raise NotImplementedError()
             
            data.append((mention, _id))

        if filter_duplicate:
            data = list(dict.fromkeys(data))
        
        # return np.array data
        data = np.array(data, dtype=object)
        
        return data

class QueryDataset_pretraining(Dataset):

    def __init__(self, data_dir, 
                filter_duplicate=False
        ):
        """       
        Parameters
        ----------
        data_dir : str
            a path of data
        filter_duplicate : bool
            filter duplicate queries
        draft : bool
            use subset of queries for debugging (default False)     
        """
        LOGGER.info("QueryDataset! data_dir={} filter_duplicate={}".format(
            data_dir,filter_duplicate
        ))
        
        self.data = self.load_data(
            data_dir=data_dir,
            filter_duplicate=filter_duplicate
        )
        
    def load_data(self, data_dir, filter_duplicate):
        """       
        Parameters
        ----------
        data_dir : str
            a path of data
        filter_composite : bool
            filter composite mentions
        filter_duplicate : bool
            filter duplicate queries  
        
        Returns
        -------
        data : np.array 
            mention, cui pairs
        """
        data = []

        #concept_files = glob.glob(os.path.join(data_dir, "*.concept"))
        with open(data_dir, "r") as f:
            lines = f.readlines()

        for row in lines:
            row = row.rstrip("\n")
            snomed_id, mention = row.split("||")
            data.append((mention, snomed_id))

        if filter_duplicate:
            data = list(dict.fromkeys(data))
        
        # return np.array data
        data = np.array(data)
        
        return data

class QueryDataset(Dataset):

    def __init__(self, data_dir, 
                filter_composite=False,
                filter_duplicate=False
        ):
        """       
        Parameters
        ----------
        data_dir : str
            a path of data
        filter_composite : bool
            filter composite mentions
        filter_duplicate : bool
            filter duplicate queries
        draft : bool
            use subset of queries for debugging (default False)     
        """
        LOGGER.info("QueryDataset! data_dir={} filter_composite={} filter_duplicate={}".format(
            data_dir, filter_composite, filter_duplicate
        ))
        
        self.data = self.load_data(
            data_dir=data_dir,
            filter_composite=filter_composite,
            filter_duplicate=filter_duplicate
        )
        
    def load_data(self, data_dir, filter_composite, filter_duplicate):
        """       
        Parameters
        ----------
        data_dir : str
            a path of data
        filter_composite : bool
            filter composite mentions
        filter_duplicate : bool
            filter duplicate queries  
        
        Returns
        -------
        data : np.array 
            mention, cui pairs
        """
        data = []

        #concept_files = glob.glob(os.path.join(data_dir, "*.txt"))
        file_types = ("*.concept", "*.txt")
        concept_files = []
        for ft in file_types:
            concept_files.extend(glob.glob(os.path.join(data_dir, ft)))

        for concept_file in tqdm(concept_files):
            with open(concept_file, "r", encoding='utf-8') as f:
                concepts = f.readlines()

            for concept in concepts:
                #print (concept)
                concept = concept.split("||")
                #if len(concept) !=5: continue
                mention = concept[3].strip().lower()
                cui = concept[4].strip()
                if cui.lower() =="cui-less": continue
                is_composite = (cui.replace("+","|").count("|") > 0)

                if filter_composite and is_composite:
                    continue
                else:
                    data.append((mention,cui))
        
        if filter_duplicate:
            data = list(dict.fromkeys(data))
        
        # return np.array data
        data = np.array(data)
        
        return data


class DictionaryDataset():
    """
    A class used to load dictionary data
    """
    def __init__(self, dictionary_path):
        """
        Parameters
        ----------
        dictionary_path : str
            The path of the dictionary
        draft : bool
            use only small subset
        """
        LOGGER.info("DictionaryDataset! dictionary_path={}".format(
            dictionary_path 
        ))
        self.data = self.load_data(dictionary_path)
        
    def load_data(self, dictionary_path):
        name_cui_map = {}
        data = []
        with open(dictionary_path, mode='r', encoding='utf-8') as f:
            lines = f.readlines()
            for line in tqdm(lines):
                line = line.strip()
                if line == "": continue
                cui, name = line.split("||")
                name = name.lower()
                if cui.lower() == "cui-less": continue
                data.append((name,cui))
        
        #LOGGER.info("concerting loaded dictionary data to numpy array...")
        #data = np.array(data)
        return data

class MetricLearningDataset_pairwise(Dataset):
    """
    Candidate Dataset for:
        query_tokens, candidate_tokens, label
    """
    def __init__(self, path, tokenizer): #d_ratio, s_score_matrix, s_candidate_idxs):
        with open(path, 'r') as f:
            lines = f.readlines()
        self.query_ids = []
        self.query_names = []
        for line in lines:
            line = line.rstrip("\n")
            query_id, name1, name2 = line.split("||")
            self.query_ids.append(query_id)
            self.query_names.append((name1, name2))
        self.tokenizer = tokenizer
        self.query_id_2_index_id = {k: v for v, k in enumerate(list(set(self.query_ids)))}
    
    def __getitem__(self, query_idx):

        query_name1 = self.query_names[query_idx][0]
        query_name2 = self.query_names[query_idx][1]
        query_id = self.query_ids[query_idx]
        query_id = int(self.query_id_2_index_id[query_id])

        return query_name1, query_name2, query_id


    def __len__(self):
        return len(self.query_names)



class MetricLearningDataset(Dataset):
    """
    Candidate Dataset for:
        query_tokens, candidate_tokens, label
    """
    def __init__(self, path, tokenizer): #d_ratio, s_score_matrix, s_candidate_idxs):
        LOGGER.info("Initializing metric learning data set! ...")
        with open(path, 'r') as f:
            lines = f.readlines()
        self.query_ids = []
        self.query_names = []
        cuis = []
        for line in lines:
            cui, _ = line.split("||")
            cuis.append(cui)

        self.cui2id = {k: v for v, k in enumerate(cuis)}
        for line in lines:
            line = line.rstrip("\n")
            cui, name = line.split("||")
            query_id = self.cui2id[cui]
            #if query_id.startswith("C"):
            #    query_id = query_id[1:]
            #query_id = int(query_id)
            self.query_ids.append(query_id)
            self.query_names.append(name)
        self.tokenizer = tokenizer
    
    def __getitem__(self, query_idx):

        query_name = self.query_names[query_idx]
        query_id = self.query_ids[query_idx]
        query_token = self.tokenizer.transform([query_name])[0]

        return torch.tensor(query_token), torch.tensor(query_id)

    def __len__(self):
        return len(self.query_names)



from itertools import groupby
from typing import List, Union
from pathlib import Path
import pandas as pd
import os

from xmen.data import features, util

from xmen.linkers import EntityLinker
from .model_wrapper import Model_Wrapper, _EMBED_DIM

from xmen.linkers.faiss_indexer import DenseFlatIndexer, DenseHNSWFlatIndexer

from xmen.log import logger

from scipy.spatial.distance import cosine


class SapBERTLinker(EntityLinker):
    """
    A class that performs entity linking using the SapBERT model.

    Args:
    - index_base_path (Union[Path, str]): Path to the directory where the index files are located.
    - kb_name (str): The name of the knowledge base to use.
    - cuda (bool): Whether to use the GPU for inference or not.
    - gpu_batch_size (int): Batch size for inference on GPU.
    - k (int): The number of closest neighbors to retrieve from the index.
    - threshold (float): A threshold value for filtering out candidate entities based on their similarity score.
    - filter_types (bool): Whether to filter out candidate entities based on their semantic type.
    - model_name (str): Name of the embedding model to use.
    - remove_duplicates (bool): Whether to remove duplicate candidate entities or not.
    - expand_abbreviations (bool): Whether to expand abbreviations or not.
    - approximate (bool): Whether to use the hierarchical index for faster inference or not.
    """

    CROSS_LINGUAL = "cambridgeltl/SapBERT-UMLS-2020AB-all-lang-from-XLMR"
    ENGLISH = "cambridgeltl/SapBERT-from-PubMedBERT-fulltext"

    # Global state
    instance = None
    model_wrapper = None
    candidate_dense_embeds = None
    term_dict = None

    # Ignore caching when using dataset.map
    def __getstate__(self):
        return {}

    @staticmethod
    def clear():
        if SapBERTLinker.instance:
            SapBERTLinker.instance.valid = False
        SapBERTLinker.instance = None
        SapBERTLinker.model_wrapper = None
        SapBERTLinker.candidate_dense_embeds = None
        SapBERTLinker.term_dict = None

    @staticmethod
    def write_index(
        index_base_path: Union[str, Path],
        term_dict: dict,
        model_name: str = CROSS_LINGUAL,
        cuda: bool = True,
        subtract_mean=True,
        batch_size=2048,
        index_buffer_size=50000,
        write_memory_map=False,
        write_flat=False,
    ):
        """
        Write index files to disk.

        Args:
        - index_base_path (Union[str, Path]): Path to the directory where the index files will be written.
        - term_dict (dict): A dictionary containing the terms to include in the index.
        - model_name (str): Name of the embedding model to use.
        - cuda (bool): Whether to use the GPU for inference or not.
        - subtract_mean (bool): Whether to subtract the mean from the embeddings or not.
        - batch_size (int): Batch size for embedding generation.
        - write_flat (bool): Whether to write the flat index to disk or not.
        """
        index_base_path = Path(index_base_path)

        index_base_path.mkdir(exist_ok=True, parents=True)

        out_dict_file = index_base_path / "dict.pickle"
        out_faiss_flat_file = index_base_path / "embed_faiss_flat.pickle"
        out_faiss_hier_file = index_base_path / "embed_faiss_hier.pickle"

        term_dict.to_pickle(out_dict_file)

        wrapper = Model_Wrapper()
        wrapper.load_model(model_name, use_cuda=cuda)

        logger.info(f"Computing dictionary embeddings with {model_name}")
        mem_map_file = (index_base_path / "embeddings.memmap") if write_memory_map else None
        if mem_map_file:
            logger.info(f"Writing embeddings to temporary memory map file {mem_map_file}")

        try:
            candidate_dense_embeds = wrapper.embed_dense(
                term_dict.term.tolist(),
                agg_mode="cls",
                use_cuda=cuda,
                show_progress=True,
                batch_size=batch_size,
                memory_map_file=mem_map_file,
            )
            if subtract_mean:
                candidate_dense_embeds -= candidate_dense_embeds.mean(0)
            # print('Writing embeddings to', out_embed_file)
            # with open(out_embed_file, "wb") as f:
            #    pickle.dump(candidate_dense_embeds, f)

            logger.info("Building FAISS Hierarchical Index")
            hier_indexer = DenseHNSWFlatIndexer(candidate_dense_embeds.shape[1], buffer_size=index_buffer_size)
            hier_indexer.index_data(candidate_dense_embeds, show_progress=True)
            logger.info(f"Writing FAISS Hierarchical index to {out_faiss_hier_file}")
            hier_indexer.serialize(str(out_faiss_hier_file))

            if write_flat:
                logger.info("Building FAISS Flat Index")
                flat_indexer = DenseFlatIndexer(candidate_dense_embeds.shape[1])
                flat_indexer.index_data(candidate_dense_embeds, show_progress=True)
                flat_indexer.serialize(str(out_faiss_flat_file))
        finally:
            if mem_map_file and mem_map_file.exists():
                os.remove(mem_map_file)

    def __init__(
        self,
        index_base_path: Union[Path, str],
        kb_name: str = "UMLS",
        cuda=True,
        gpu_batch_size: int = 16,
        k: int = 10,
        threshold: float = 0.0,
        model_name: str = CROSS_LINGUAL,
        remove_duplicates=True,
        expand_abbreviations=True,
        approximate=True,
    ):
        index_base_path = Path(index_base_path)
        term_dict_pkl = index_base_path / "dict.pickle"

        if SapBERTLinker.instance:
            raise Exception("SapBERTLinker is a singleton")
        SapBERTLinker.model_wrapper = Model_Wrapper()
        SapBERTLinker.model_wrapper.load_model(model_name, use_cuda=cuda)
        self.cuda = cuda
        if approximate:
            logger.info("Loading hierarchical faiss index")
            self.indexer = DenseHNSWFlatIndexer(_EMBED_DIM)
            self.indexer.deserialize_from(str(index_base_path / "embed_faiss_hier.pickle"))
        else:
            logger.info("Loading flat faiss index")
            self.indexer = DenseFlatIndexer(_EMBED_DIM)
            self.indexer.deserialize_from(str(index_base_path / "embed_faiss_flat.pickle"))
        # with open(dict_embeddings_pkl, "rb") as f:
        #    get_logger().info("Loading embeddings")
        #    SapBERTLinker.candidate_dense_embeds = pickle.load(f)
        with open(term_dict_pkl, "rb") as f:
            SapBERTLinker.term_dict = pd.read_pickle(f)
            self.term_dict_idx = SapBERTLinker.term_dict.set_index("cui")
        self.k = k
        self.kb_name = kb_name
        self.threshold = threshold
        self.gpu_batch_size = gpu_batch_size
        self.remove_duplicates = remove_duplicates
        self.expand_abbreviations = expand_abbreviations

        SapBERTLinker.instance = self
        self.valid = True

    def predict_batch(self, dataset, batch_size=1):
        """
        Perform entity linking on a batch of sentences.

        Args:
        - dataset (Dataset): A `Dataset` object containing the sentences to link entities in.
        - batch_size (int): The batch size to use for linking.

        Returns:
        - List of lists of tuples: A list of lists of tuples. Each tuple contains information about a linked entity,
        - including its URI, label, score, and span in the sentence.

        Raises:
        - Exception: If `SapBERTLinker` instance has been cleared, the linker instance is in an inconsistent state.
        """
        if not self.valid:
            raise Exception("SapBERT instance has been cleared, this linker instance is in an inconsistent state.")
        expand_abbreviations = self.expand_abbreviations

        def get_result(sample):
            def get_str(mention):
                s = " ".join(mention["text"])
                if expand_abbreviations and mention.get("long_form"):
                    s += " [SEP] " + mention["long_form"]
                return s

            sample = util.init_schema(sample)
            entities = sample["entities"]

            mentions = tuple(
                zip(*[(j, get_str(mention)) for j, doc_entities in enumerate(entities) for mention in doc_entities])
            )
            if not mentions:  # empty
                return {"entities": entities}

            mentions_index, mention_strings = mentions

            concepts = self._get_concepts(list(mention_strings))

            for mi, concept_group in groupby(zip(mentions_index, concepts), key=lambda p: p[0]):
                for j, c in enumerate(concept_group):
                    entities[mi][j]["normalized"] = c[1]

            return {"entities": entities}

        return dataset.map(
            function=get_result,
            batch_size=batch_size,
            batched=True,
            load_from_cache_file=False,
            features=features,
        )

    def predict(self, unit: str, entities: dict) -> dict:
        raise NotImplementedError()

    def _get_concepts(self, mention_candidates):
        logger.debug(f"Calculate embeddings for {len(mention_candidates)} mentions")
        mention_dense_embeds = SapBERTLinker.model_wrapper.embed_dense(
            mention_candidates,
            show_progress=False,
            agg_mode="cls",
            use_cuda=self.cuda,
            batch_size=self.gpu_batch_size,
        )
        scores, candidate_idxs_batch, vectors = self.indexer.search_knn(mention_dense_embeds, self.k)

        for i, embed in enumerate(mention_dense_embeds):
            candidates = candidate_idxs_batch[i]
            top_scores = scores[i]
            vector = vectors[i]
            top_scores = [1 - cosine(embed, v[0:_EMBED_DIM]) for v in vector]
            concepts = []
            cuis = set()
            for r, score in zip(SapBERTLinker.term_dict.iloc[candidates].iterrows(), top_scores):
                r = r[1]
                if score > self.threshold:
                    if not self.remove_duplicates or not r.cui in cuis:
                        cuis.add(r.cui)
                        concepts.append((r.cui, score))
            yield [
                {"db_id": cui, "score": score, "db_name": self.kb_name, "predicted_by": ["sapbert"]}
                for cui, score in sorted(concepts, key=lambda p: -p[1])
            ]



from abc import ABC, abstractmethod
from datasets.arrow_dataset import Dataset
import numpy as np


class Reranker(ABC):
    @abstractmethod
    def rerank_batch(self, dataset: Dataset) -> Dataset:
        pass

    @staticmethod
    def sort_concepts(concepts):
        idx = np.argsort(np.array(concepts["score"]))[-1::-1]
        result = {}
        for k, v in concepts.items():
            result[k] = list(np.array(v)[idx])
        return result


from .cross_encoder import CrossEncoderReranker



from typing import Iterable
from itertools import groupby
import pandas as pd
import warnings
import regex as re

from .ext.neleval.prepare import SelectAlternatives
from .ext.neleval.document import Document
from .ext.neleval.evaluate import Evaluate, StrictMetricWarning
from .ext.neleval.annotation import Annotation, Candidate
from .ext.neleval.configs import get_measure

# constants for interacting with different neleval measures
_PARTIAL_EVAL_MEASURE_FMT_STRING = "overlap-maxmax:None:span+kbid"

# equivalent to strong_all_match
_STRICT_EVAL_MEASURE_FMT_STRING = "sets:None:span+kbid"

_NER_STRICT_EVAL_MEASURE_FMT_STRING = "sets:None:span"

# constants for interacting with different neleval measures
_NER_PARTIAL_EVAL_MEASURE_FMT_STRING = "overlap-maxmax:None:span"

# Any match within the unit / sentence counts
_LOOSE_EVAL_MEASURE_FMT_STRING = "sets:None:docid+kbid"

_KEY_TO_METRIC = {
    "strict": _STRICT_EVAL_MEASURE_FMT_STRING,
    "partial": _PARTIAL_EVAL_MEASURE_FMT_STRING,
    "loose": _LOOSE_EVAL_MEASURE_FMT_STRING,
    "ner_strict": _NER_STRICT_EVAL_MEASURE_FMT_STRING,
    "ner_partial": _NER_PARTIAL_EVAL_MEASURE_FMT_STRING,
}


def _get_word_len(row):
    if row.gt_text:
        return len(" ".join(row.gt_text).split(" "))
    else:
        return None


def _get_entity_info(row):
    res = {}
    res["_word_len"] = _get_word_len(row)
    res["_abbrev"] = (len(row.gt_text) == 1) and bool(re.match("[A-Z]{2,3}", row.gt_text[0])) if row.gt_text else None
    return res


def error_analysis(
    ground_truth: Iterable, prediction: Iterable, tasks: list = ["ner", "nen"], allow_multiple_gold_candidates=False
) -> pd.DataFrame:
    """
    Computes error analysis of entity linking predictions by comparing against the ground truth entities, assuming that the entities are aligned.

    Args:
    - ground_truth: An iterable of dictionaries representing the ground truth entities for each document.
    - prediction: An iterable of dictionaries representing the predicted entities for each document.
    - tasks: list of either 'ner' (for named entity recogition), 'nen' (for entity linking), or both
    - allow_multiple_gold_candidates: A boolean indicating whether multiple ground truth entities per mention are allowed. Defaults to False.

    Returns:
    - A pandas DataFrame containing error analysis of entity linking predictions.
    """

    res = []
    for gt, pred in zip(ground_truth, prediction):
        error_df = _get_error_df(gt["entities"], pred["entities"], tasks, allow_multiple_gold_candidates)
        if "corpus_id" in gt:
            error_df["corpus_id"] = gt["corpus_id"]
        error_df["document_id"] = gt["document_id"]
        res.append(error_df)
    ea_df = pd.concat(res).reset_index().drop(columns="index")
    if len(ea_df) == 0:
        return ea_df
    entity_info = pd.DataFrame(list(ea_df.apply(_get_entity_info, axis=1)))
    return pd.concat([entity_info, ea_df], axis=1)


def _get_error_df(gt_ents, pred_ents, tasks, allow_multiple_gold_candidates) -> pd.DataFrame:
    """
    Construct a Pandas DataFrame with the error analysis results from the comparison of two lists of named entities.

    Args:
    - gt_ents (list): A list of dictionaries, each representing a ground truth named entity.
    - pred_ents (list): A list of dictionaries, each representing a predicted named entity. Each dictionary must have the same keys as in 'gt_ents'.
    - tasks: list of either 'ner' (for named entity recogition), 'nen' (for entity linking), or both
    - allow_multiple_gold_candidates (bool): A boolean flag indicating whether to allow multiple ground truth entities to match a single predicted entity. Defaults to False.

    Returns:
    A Pandas DataFrame with the errors.
    """

    def get_items(entities):
        return [
            (
                e["offsets"][0][0],
                e["offsets"][-1][1],
                e["normalized"] if len(e["normalized"]) > 0 else [{"db_id": "NIL"}],
                e["text"],
                e["type"],
            )
            for e in sorted(entities, key=lambda e: (e["offsets"], e["type"], e["text"]))
        ]

    gt_items = get_items(gt_ents)
    pred_items_unmatched = get_items(pred_ents)

    # If we have entities with multiple normalizations, we want to order by best match order
    pred_items = []
    for key, group in groupby(pred_items_unmatched, lambda x: (x[0], x[1], x[3], x[4])):
        group = list(group)
        if len(group) == 1:
            pred_items.append(group[0])
        else:

            def best_match_index(pred_normalized):
                matches = [g for g in gt_items if (g[0], g[1], g[3], g[4]) == key]
                best_match = len(pred_normalized)
                best_index = len(pred_normalized)
                for i, m in enumerate(matches):
                    for j, p in enumerate(pred_normalized):
                        if p["db_id"] == m[2][0]["db_id"]:
                            if j < best_index:
                                best_index = j
                                best_match = i
                return best_match

            best_match_indices = sorted([(pred, best_match_index(pred[2])) for pred in group], key=lambda t: t[1])
            matched = [None] * len(best_match_indices)
            for p, i in best_match_indices:
                if i < len(matched) and matched[i] == None:
                    matched[i] = p
                else:
                    found = False
                    for j in range(len(matched)):
                        if not found and matched[j] == None:
                            matched[j] = p
                            found = True
                    assert found == True

            pred_items.extend(matched)
    assert len(pred_items) == len(pred_items_unmatched)

    ent_res = []

    def record_match(
        pred_s: int, pred_e: int, pred_c, pred_text, pred_type, gt_s, gt_e, gt_c, gt_text, gt_type, e_match_type
    ):
        if (not gt_c and pred_c) or e_match_type == "fp":  # false positive
            ent_res.append(
                {
                    "pred_start": pred_s,
                    "pred_end": pred_e,
                    "pred_text": pred_text,
                    "gt_start": None,
                    "gt_end": None,
                    "gt_text": None,
                    "ner_match_type": e_match_type,
                    "gold_concept": None,
                    "gold_type": None,
                    "pred_index": -1,
                    "pred_index_score": None,
                    "pred_top": None,
                    "pred_top_score": None,
                }
            )
            return

        def get_match_result(gt_concept):
            is_not_fn = e_match_type != "fn"
            pred_ids = [c["db_id"] for c in pred_c]
            idx = pred_ids.index(gt_concept["db_id"]) if gt_concept["db_id"] in pred_ids else -1
            return {
                "pred_start": pred_s if is_not_fn else None,
                "pred_end": pred_e if is_not_fn else None,
                "pred_text": pred_text if is_not_fn else None,
                "gt_start": gt_s,
                "gt_end": gt_e,
                "gt_text": gt_text,
                "ner_match_type": e_match_type,
                "gold_concept": gt_concept,
                "gold_type": gt_type,
                "pred_index": int(idx),
                "pred_index_score": pred_c[idx].get("score", None) if idx >= 0 else None,
                "pred_top": pred_c[0]["db_id"] if len(pred_c) > 0 else None,
                "pred_top_score": pred_c[0].get("score", None) if len(pred_c) > 0 else None,
            }

        if not pred_c and gt_c:
            assert len(gt_c) == 1
            ent_res.append(
                {
                    "pred_start": pred_s,
                    "pred_end": pred_e,
                    "pred_text": pred_text,
                    "gt_start": gt_s,
                    "gt_end": gt_e,
                    "gt_text": gt_text,
                    "ner_match_type": e_match_type,
                    "gold_concept": gt_c[0],
                    "gold_type": "",
                    "pred_index": -1,
                    "pred_index_score": None,
                    "pred_top": None,
                    "pred_top_score": None,
                }
            )
            return

        if allow_multiple_gold_candidates:
            matches = [get_match_result(gt_concept) for gt_concept in gt_c]
            true_matches = [m["pred_index"] for m in matches if m["pred_index"] != -1]
            if true_matches:
                best_idx = min(true_matches)
                best_matches = [m for m in matches if m["pred_index"] == best_idx]
                assert len(best_matches) == 1
                ent_res.append(best_matches[0])
            else:  # No matches for any of the concepts
                ent_res.append(matches[0])

        else:
            for gt_concept in gt_c:
                ent_res.append(get_match_result(gt_concept))

    # Initialize variables
    gt_s = gt_e = gt_text = gt_type = None
    pred_s = pred_e = pred_text = pred_type = None
    pred_c = gt_c = []

    while gt_items or pred_items:
        if not gt_items:
            e_match_type = "fp"
            pred_s, pred_e, pred_c, pred_text, pred_type = pred_items.pop()
            record_match(
                pred_s,
                pred_e,
                pred_c,
                pred_text,
                pred_type,
                -1,
                -1,
                None,
                None,
                None,
                e_match_type,
            )
        elif not pred_items:
            e_match_type = "fn"
            gt_s, gt_e, gt_c, gt_text, gt_type = gt_items.pop()
            record_match(
                pred_s,
                pred_e,
                pred_c,
                pred_text,
                pred_type,
                gt_s,
                gt_e,
                gt_c,
                gt_text,
                gt_type,
                e_match_type,
            )
        else:
            pred_s, pred_e, pred_c, pred_text, pred_type = pred_items[0]
            gt_s, gt_e, gt_c, gt_text, gt_type = gt_items[0]

            if pred_s == gt_s and pred_e == gt_e:
                e_match_type = "tp"
                record_match(
                    pred_s,
                    pred_e,
                    pred_c,
                    pred_text,
                    pred_type,
                    gt_s,
                    gt_e,
                    gt_c,
                    gt_text,
                    gt_type,
                    e_match_type,
                )
                gt_items.pop(0)
                pred_items.pop(0)
            elif pred_s >= gt_e:
                e_match_type = "fn"
                record_match(
                    pred_s,
                    pred_e,
                    pred_c,
                    pred_text,
                    pred_type,
                    gt_s,
                    gt_e,
                    gt_c,
                    gt_text,
                    gt_type,
                    e_match_type,
                )
                gt_items.pop(0)
            elif pred_e <= gt_s:
                e_match_type = "fp"
                record_match(
                    pred_s,
                    pred_e,
                    pred_c,
                    pred_text,
                    pred_type,
                    gt_s,
                    gt_e,
                    gt_c,
                    gt_text,
                    gt_type,
                    e_match_type,
                )
                pred_items.pop(0)
            else:
                e_match_type = "be"
                pred_s, pred_e, pred_c, pred_text, pred_type = pred_items.pop(0)
                gt_s, gt_e, gt_c, gt_text, gt_type = gt_items.pop(0)
                while True:
                    record_match(
                        pred_s,
                        pred_e,
                        pred_c,
                        pred_text,
                        pred_type,
                        gt_s,
                        gt_e,
                        gt_c,
                        gt_text,
                        gt_type,
                        e_match_type,
                    )
                    if pred_e < gt_e:
                        if pred_items and pred_items[0][0] <= gt_e:
                            pred_s, pred_e, pred_c, pred_text, pred_type = pred_items.pop(0)
                        else:
                            # if gt_items: Why?
                            #    gt_items.pop(0)
                            break
                    elif gt_e < pred_e:
                        if gt_items and gt_items[0][0] <= pred_e:
                            gt_s, gt_e, gt_c, gt_text, gt_type = gt_items.pop(0)
                        else:
                            # if pred_items: Why?
                            #    pred_items.pop(0)
                            break
                    else:
                        break

    res_cols = ["gt_start", "gt_end", "gt_text"]
    if "ner" in tasks:
        res_cols += ["pred_start", "pred_end", "pred_text", "ner_match_type"]
    if "nen" in tasks:
        res_cols += ["gold_concept", "gold_type", "pred_index", "pred_index_score", "pred_top", "pred_top_score"]

    if ent_res:
        return pd.DataFrame(ent_res)[res_cols]
    else:
        return pd.DataFrame(columns=res_cols)


def get_synset(kb, scui):
    kb_ix = kb.cui_to_entity[str(scui)]
    return set([kb_ix.canonical_name] + kb_ix.aliases)


def evaluate(
    ground_truth: Iterable,
    prediction: Iterable,
    allow_multiple_gold_candidates=False,
    top_k_predictions=1,
    threshold=None,
    ner_only=False,
    metrics=["strict"],
) -> dict:
    """
    Evaluate the performance of the system.

    Args:
    - ground_truth: An iterable of ground truth annotations.
    - prediction: An iterable of predicted annotations.
    - allow_multiple_gold_candidates: A boolean indicating whether multiple
            gold standard candidates are allowed for each annotation. Defaults to False.
    - top_k_predictions: An integer indicating the maximum number of predicted
            candidates to consider for each annotation. Defaults to 1.
    - threshold: A float indicating the minimum confidence threshold to consider
            for each predicted candidate. Defaults to None.
    - ner_only: A boolean indicating whether to only consider NER tags. Defaults to False.
    - metrics: A list of strings indicating which metrics to use for evaluation. Defaults to ["strict"].

    Returns:
    - A dictionary containing the evaluation results for each specified metric.
    """
    with warnings.catch_warnings():
        # Ignore division by zero problems raised by neleval
        warnings.filterwarnings("ignore", category=StrictMetricWarning)
        if ner_only:
            allow_multiple_gold_candidates = False
        system_docs = list(
            _to_nel_eval(
                prediction,
                allow_multiple_candidates=False,
                top_k=top_k_predictions,
                threshold=threshold,
                ner_only=ner_only,
            )
        )
        gold_docs = list(
            _to_nel_eval(
                ground_truth,
                allow_multiple_candidates=allow_multiple_gold_candidates,
                top_k=None,
                threshold=None,
                ner_only=ner_only,
            )
        )

        num_annos_system = sum([len(a.candidates) for d in system_docs for a in d.annotations])
        num_annos_gold = sum([len(a.candidates) for d in gold_docs for a in d.annotations])

        if allow_multiple_gold_candidates:
            SelectAlternatives(system_docs, gold_docs)()

        if metrics == "all":
            metrics = list(_KEY_TO_METRIC.keys())

        eval_fn = Evaluate(
            system=system_docs,
            gold=gold_docs,
            measures=[get_measure(_KEY_TO_METRIC[m]) for m in metrics],
        )
        res = eval_fn()
        for v in res.values():
            v["n_docs_system"] = len(system_docs)
            v["n_annos_system"] = num_annos_system
            v["n_docs_gold"] = len(gold_docs)
            v["n_annos_gold"] = num_annos_gold

        return {k: res[_KEY_TO_METRIC[k]] for k in metrics}


def _to_nel_eval(
    units: Iterable,
    allow_multiple_candidates: bool = False,
    top_k: int = None,
    threshold: float = None,
    ner_only=False,
) -> Iterable:
    """
    Converts a list of units to a list of Document objects in neleval format.

    Args:
    - units: An iterable of units containing entity annotations.
    - allow_multiple_candidates: A boolean indicating whether to allow multiple candidate entities.
    - top_k: An integer indicating the number of top candidates to keep.
    - threshold: A float indicating the score threshold for keeping candidates.
    - ner_only: A boolean indicating whether to ignore entity types.

    Returns:
    - An iterable of Document objects in neleval format.
    """
    for u in units:
        entities = sorted(u["entities"], key=lambda e: e["offsets"])
        cid = u["corpus_id"] if "corpus_id" in u and u["corpus_id"] else ""
        unit_id = cid + u["document_id"]
        annotations = []
        for e in entities:
            if "normalized" in e:
                for c in e["normalized"]:
                    if not "type" in c:
                        c["type"] = None
                    if not "score" in c:
                        c["score"] = None
            else:
                e["normalized"] = []
            start, end = min([o[0] for o in e["offsets"]]), max([o[1] for o in e["offsets"]])
            if ner_only:
                anno = Annotation(unit_id, start, end, [])
                annotations.append(anno)
            else:
                candidates = []
                for c in e["normalized"]:
                    if not threshold or not c["score"] or c["score"] >= threshold:
                        candidates.append(Candidate(c["db_id"], c["score"], c["type"]))
                candidates.sort(key=lambda c: -c.score if c.score else 0)
                if top_k:
                    candidates = candidates[:top_k]
                if allow_multiple_candidates:
                    anno = Annotation(unit_id, start, end, candidates)
                    annotations.append(anno)
                else:
                    for cand in candidates:
                        anno = Annotation(unit_id, start, end, [cand])
                        annotations.append(anno)
        yield Document(unit_id, annotations)


def evaluate_at_k(ground_truth, pred, eval_k=[1, 2, 4, 8, 16, 32, 64], silent=False):
    """
    Runs evaluation for different values of k (number of candidates)

    Args:
    - ground_truth: ground truth annotations
    - pred: predicted candidates
    - eval_k: different values of k to evaluate
    - silent: whether to print results to standard output

    Returns:
    - a dictionary of evaluation results for each k
    """
    res = {}
    for ki in eval_k:
        eval_res = evaluate(ground_truth, pred, top_k_predictions=ki)
        if not silent:
            print(f"Recall@{ki}", eval_res["strict"]["recall"])
        res[ki] = eval_res
    return res



from typing import Union, List


def init_schema(dataset):
    """
    Initializes the non-BigBIO elements of a dataset
    """
    n_docs = len(dataset["document_id"])
    if not "corpus_id" in dataset:
        dataset["corpus_id"] = [None] * n_docs
    if not "lang" in dataset:
        dataset["lang"] = [None] * n_docs
    for es in dataset["entities"]:
        for e in es:
            for n in e["normalized"]:
                if not "score" in n:
                    n["score"] = 0.0
                if not "predicted_by" in n:
                    n["predicted_by"] = []
            if not "long_form" in e:
                e["long_form"] = None
    return dataset


class Concept:
    """
    A class representing a concept.

    Args:
    - db_id (str): The id of the concept.
    - score (float): The score of the concept.
    - db_name (str): The name of the knowledge base containing the concept.
    - type (str): The type of the concept.

    Attributes:
    - _dict (dict): A dictionary containing the 'db_id', 'target_kb', 'type', and 'score' keys.
    """

    def __init__(
        self,
        db_id: str = None,
        score: float = None,
        db_name: str = None,
        type: str = None,
    ):
        self._dict = {
            "db_id": db_id,
            "target_kb": db_name,
            "type": type,
            "score": score,
        }


class Entity:
    """
    A class representing an entity.

    Args:
    - offsets: A tuple or list of tuples representing the start and end character offsets of the entity.
    - text (Union[str, List[str]]): A string or list of strings representing the text of the entity.
    - id (str): The id of the entity.
    - entity_type (str): The type of the entity.
    - concepts (List[Concept]): A list of concepts associated with the entity.

    Attributes:
    - _dict (dict): A dictionary containing the 'id', 'text', 'offsets', 'type', and 'normalized' keys.
    """

    def __init__(
        self,
        offsets,
        text: Union[str, List[str]],
        id: str = "1",
        entity_type: str = None,
        concepts: List[Concept] = None,
    ):
        self._dict = {
            "id": id,
            "text": [text] if type(text) == str else text,
            "offsets": offsets,
            "type": entity_type,
            "normalized": [c._dict for c in concepts] if concepts else [],
        }


def make_document(entities: List[Entity], document_id: str = "1", corpus_id: str = "x") -> dict:
    """
    Creates a dictionary representing a document.

    Args:
    - entities (List[Entity]): A list of entities in the document.
    - document_id (str): The id of the document.
    - corpus_id (str): The id of the corpus containing the document.

    Returns:
    - A dictionary with the 'corpus_id', 'document_id', and 'entities' keys.
    """
    return {
        "corpus_id": corpus_id,
        "document_id": document_id,
        "entities": [e._dict for e in entities],
    }


def get_cuis(dataset):
    """
    Extracts the unique CUIs (Concept Unique Identifiers) from a dataset.

    Args:
    - dataset: A dictionary containing the 'entities' key, which is a list of Entity objects.

    Returns:
    - A list of unique CUIs.
    """
    return [c["db_id"] for doc in dataset["entities"] for e in doc for c in e["normalized"]]


def filter_and_apply_threshold(input_pred, k: int, threshold: float):
    """
    Filters the predicted entities based on a score threshold and returns the top k entities for each entity.

    Args:
    - input_pred (list): A list of predicted entities to be filtered and thresholded.
    - k (int): The number of top entities to keep after filtering.
    - threshold (float): The score threshold below which entities will be removed.

    Returns:
    - list: A list of filtered and thresholded predicted entities.
    """
    assert k >= 0

    def apply(entry):
        entities = entry["entities"]
        for e in entities:
            filtered = [n for n in e["normalized"] if "score" in n and n["score"] >= threshold]
            e["normalized"] = sorted(filtered, key=lambda n: n["score"], reverse=True)[:k]
        return {"entities": entities}

    return input_pred.map(apply, load_from_cache_file=False)



from xmen.umls import get_umls_concepts
import importlib.util
import sys
import json
from omegaconf import DictConfig
from .utils import get_alias_count
from ..log import logger


def assemble(cfg: DictConfig, output: str, custom_path=None):
    """
    Assembles a list of concept details into a JSONL file at the given output path.

    Args:
    - cfg: The configuration dictionary.
    - output: The output path for the JSONL file.
    - custom_path: The path to custom code, if applicable.
    """
    concept_details = get_concept_details(cfg, custom_path)

    n_concepts = len(concept_details)
    if n_concepts == 0:
        logger.warn("List of concepts is empty")

    logger.info(f"Number of concepts: {n_concepts}")
    logger.info(f"Number of aliases: {get_alias_count(concept_details)}")

    output.parent.mkdir(exist_ok=True, parents=True)
    with open(output, "w") as fout:
        for value in concept_details.values():
            fout.write(json.dumps(value) + "\n")


def get_concept_details(cfg, custom_path):
    """
    Returns the concept details from UMLS or custom code.

    Args:
    - cfg: The OmegaConf config object.
    - custom_path (str): The path to the custom code file.

    Returns:
    - dict: A dictionary containing the concept details.
    """
    if "umls" in cfg.dict:
        return get_umls_concepts(
            cfg.dict.umls.meta_path,
            cfg.dict.umls.get("lang", []),
            sabs=cfg.dict.umls.get("sabs", []),
            sources=cfg.dict.umls.get("sources", []),
            semantic_groups=cfg.dict.umls.get("semantic_groups", None),
            semantic_types=cfg.dict.umls.get("semantic_types", None),
            non_suppressed_only=cfg.dict.umls.get("non_suppressed_only", False),
            semantic_group_file_version=cfg.dict.umls.get("semantic_group_file_version", "v03"),
        )
    elif "custom" in cfg.dict:
        return get_custom_concepts(cfg, custom_path)


def get_custom_concepts(cfg, path) -> dict:
    """
    Load and execute a Python module located in the given path that implements a function `get_concept_details(cfg)`.
    This function should return a dictionary containing details about custom concepts.

    Args:
    - cfg: Configuration object to pass to the `get_concept_details` function.
    - path: Path to the Python module containing the `get_concept_details` function.

    Returns:
    - Dictionary containing details about custom concepts.
    """
    spec = importlib.util.spec_from_file_location("custom_code", path)
    custom_module = importlib.util.module_from_spec(spec)
    sys.modules["custom_code"] = custom_module
    spec.loader.exec_module(custom_module)
    return custom_module.get_concept_details(cfg)



# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#
"""
FAISS-based index components. Original from 
https://github.com/facebookresearch/BLINK/blob/main/blink/indexer/faiss_indexer.py
"""

import os
import logging
import pickle

import faiss
import numpy as np

from xmen.log import logger
from tqdm.autonotebook import tqdm


class DenseIndexer(object):
    """
    A class for creating and searching a dense vector index using Faiss.

    Args:
    - buffer_size (int): The maximum number of data points to hold in memory while indexing.

    Attributes:
    - buffer_size (int): The maximum number of data points to hold in memory while indexing.
    - index_id_to_db_id (list): A list that maps the index ID to the original database ID.
    - index (faiss.IndexFlatIP): The Faiss index.
    """

    def __init__(self, buffer_size: int = 50000):
        self.buffer_size = buffer_size
        self.index_id_to_db_id = []
        self.index = None

    def index_data(self, data: np.array, show_progress=False):
        raise NotImplementedError

    def search_knn(self, query_vectors: np.array, top_docs: int):
        raise NotImplementedError

    def serialize(self, index_file: str):
        """
        Serializes the index to a file.

        Args:
        - index_file (str): The path to the file where the serialized index will be stored.
        """
        logger.info("Serializing index to %s", index_file)
        faiss.write_index(self.index, index_file)

    def deserialize_from(self, index_file: str):
        """
        Deserializes the index from a file.

        Args:
        - index_file (str): The path to the file containing the serialized index.
        """
        logger.info("Loading index from %s", index_file)
        self.index = faiss.read_index(index_file)
        logger.info("Loaded index of type %s and size %d", type(self.index), self.index.ntotal)


# DenseFlatIndexer does exact search
class DenseFlatIndexer(DenseIndexer):
    """
    Represents a dense flat indexer object.

    Args:
    - vector_sz (int): Size of the input vector. Default is 1.
    - buffer_size (int): Size of the buffer to use while indexing data. Default is 50000.

    Attributes:
    - index: An instance of faiss.IndexFlatIP initialized with vector_sz.
    """

    def __init__(self, vector_sz: int = 1, buffer_size: int = 50000):
        super(DenseFlatIndexer, self).__init__(buffer_size=buffer_size)
        self.index = faiss.IndexFlatIP(vector_sz)

    def index_data(self, data: np.array, show_progress=False):
        """
        Indexes the given data using the IndexFlatIP index.

        Args:
        - data (np.array): The input data to be indexed.
        - show_progress (bool): If True, shows progress bar while indexing. Default is False.
        """
        n = len(data)
        # indexing in batches is beneficial for many faiss index types
        logger.info("Indexing data, this may take a while.")
        cnt = 0
        for i in tqdm(range(0, n, self.buffer_size), disable=not show_progress):
            vectors = [np.reshape(t, (1, -1)) for t in data[i : i + self.buffer_size]]
            vectors = np.concatenate(vectors, axis=0)
            self.index.add(vectors)
            cnt += self.buffer_size

        logger.info("Total data indexed %d", n)

    def search_knn(self, query_vectors, top_k):
        """
        Searches the indexed data for the nearest neighbors of the given query vectors.

        Args:
        - query_vectors (np.array): The query vectors for which nearest neighbors are to be found.
        - top_k (int): The number of nearest neighbors to be returned.

        Returns:
        - scores (np.array): The similarity scores of the nearest neighbors.
        - indexes (np.array): The indexes of the nearest neighbors in the indexed data.
        """
        scores, indexes = self.index.search(query_vectors, top_k)
        return scores, indexes


# DenseHNSWFlatIndexer does approximate search
class DenseHNSWFlatIndexer(DenseIndexer):
    """
    Efficient index for retrieval using HNSWFlat algorithm with L2 similarity.
    This indexer supports conversion from DOT product similarity space to L2 similarity space.
    Default settings are for high accuracy but also high RAM usage.

    Args:
    - vector_sz (int): Size of input vectors.
    - buffer_size (int): Batch size for indexing. Default is 50000.
    - store_n (int): Number of vectors to store per index. Default is 128.
    - ef_search (int): Number of neighbors to consider during search. Default is 256.
    - ef_construction (int): Number of neighbors to consider during index construction. Default is 200.

    Attributes:
    - index: Faiss IndexHNSWFlat object for indexing vectors.
    - phi (float): Auxiliary dimension for conversion of DOT product similarity space to L2 similarity space.
    """

    def __init__(
        self,
        vector_sz: int,
        buffer_size: int = 50000,
        store_n: int = 128,
        ef_search: int = 256,
        ef_construction: int = 200,
    ):
        super(DenseHNSWFlatIndexer, self).__init__(buffer_size=buffer_size)

        # IndexHNSWFlat supports L2 similarity only
        # so we have to apply DOT -> L2 similairy space conversion with the help of an extra dimension
        index = faiss.IndexHNSWFlat(vector_sz + 1, store_n)
        index.hnsw.efSearch = ef_search
        index.hnsw.efConstruction = ef_construction
        self.index = index
        self.phi = 0

    def index_data(self, data: np.array, show_progress=False):
        """
        Indexes the input data into the indexer.

        Args:
        - data (np.array): Input data to be indexed.
        - show_progress (bool): If True, shows a progress bar during indexing. Default is False.
        """
        n = len(data)

        # max norm is required before putting all vectors in the index to convert inner product similarity to L2
        if self.phi > 0:
            raise RuntimeError(
                "DPR HNSWF index needs to index all data at once," "results will be unpredictable otherwise."
            )
        phi = 0
        for i, item in enumerate(data):
            doc_vector = item
            norms = (doc_vector**2).sum()
            phi = max(phi, norms)
        logger.info("HNSWF DotProduct -> L2 space phi={}".format(phi))
        self.phi = 0

        # indexing in batches is beneficial for many faiss index types
        logger.info("Indexing data, this may take a while.")
        cnt = 0
        for i in tqdm(range(0, n, self.buffer_size), disable=not show_progress):
            vectors = [np.reshape(t, (1, -1)) for t in data[i : i + self.buffer_size]]

            norms = [(doc_vector**2).sum() for doc_vector in vectors]
            aux_dims = [np.sqrt(phi - norm) for norm in norms]
            hnsw_vectors = [np.hstack((doc_vector, aux_dims[i].reshape(-1, 1))) for i, doc_vector in enumerate(vectors)]
            hnsw_vectors = np.concatenate(hnsw_vectors, axis=0)

            self.index.add(hnsw_vectors)
            cnt += self.buffer_size

        logger.info("Total data indexed %d" % n)

    def search_knn(self, query_vectors, top_k):
        """
        Searches the index for the k-nearest neighbors of the query vectors.

        Args:
        - query_vectors: A numpy array of shape (n_queries, embedding_dim) containing the query vectors.
        - top_k: An integer representing the number of nearest neighbors to retrieve for each query.

        Returns:
        - scores: A numpy array of shape (n_queries, top_k) containing the distances of the k-nearest neighbors for each query.
        - indexes: A numpy array of shape (n_queries, top_k) containing the indexes of the k-nearest neighbors for each query.
        - vectors: A numpy array of shape (n_queries, top_k, embedding_dim) containing the embeddings of the k-nearest neighbors for each query.
        """
        aux_dim = np.zeros(len(query_vectors), dtype="float32")
        query_nhsw_vectors = np.hstack((query_vectors, aux_dim.reshape(-1, 1)))
        logger.debug("query_hnsw_vectors %s", query_nhsw_vectors.shape)
        scores, indexes, vectors = self.index.search_and_reconstruct(query_nhsw_vectors, top_k)
        return scores, indexes, vectors

    def deserialize_from(self, file: str):
        """
        Deserializes the index from the given file.

        Args:
        - file: A string representing the path to the file containing the serialized index.
        """
        super(DenseHNSWFlatIndexer, self).deserialize_from(file)
        # to trigger warning on subsequent indexing
        self.phi = 1



from itertools import groupby


class ConceptMerger:
    """
    A class to merge entities in text data based on their text and offset.
    """

    def transform_batch(self, dataset):
        """
        Merges entities in a given dataset based on their text and offset.

        Args:
        - dataset (Dataset): A dataset object from the Hugging Face datasets library.
        """

        def _merge_entities(d):
            ents = d["entities"]
            merged = []
            for k, grp in groupby(
                sorted(ents, key=lambda e: e["offsets"][0][0]),
                lambda e: (e["text"], e["offsets"]),
            ):
                grp = list(grp)
                normalized = []
                for e in grp:
                    for n in e["normalized"]:
                        if not n in normalized:
                            normalized.append(n)
                merged.append(
                    {
                        "id": "+".join([e["id"] for e in grp]),
                        "normalized": normalized,
                        "type": [e["type"] for e in grp],
                        "text": k[0],
                        "offsets": k[1],
                    }
                )
            return {"entities": merged}

        return dataset.map(_merge_entities)



"Adapted from https://github.com/wikilinks/neleval/blob/master/neleval/ for Python >3.10 compatibility"

from collections import defaultdict
import json

from .document import Reader
from neleval.utils import utf8_open


class SelectAlternatives(object):
    """Handle KB ambiguity in the gold standard by modifying it to match system

    The following back-off strategy applies for each span with gold standard
    ambiguity:

        * attempt to match it to the top candidate for that span
        * attempt to match it to the top candidate for any span in that
          document
        * attempt to match it to the top candidate for any span in the
          collection
        * default to select the first listed candidate

    The altered gold standard will be output.
    """

    def __init__(self, system, gold, fields='eid'):
        self.system = system
        self.gold = gold
        self.fields = fields.split(',') if fields != '*' else '*'

    def _get_key(self, candidate):
        if self.fields == '*':
            # avoid comparing on score
            return (candidate.eid, candidate.__dict__)
        return tuple(getattr(candidate, field, None)
                     for field in self.fields)

    def __call__(self):
        system = self.system
        if not isinstance(system, list):
            system = Reader(utf8_open(system))
        gold = self.gold
        if not isinstance(gold, list):
            gold = Reader(utf8_open(gold))

        # XXX: maybe we should choose most frequent rather than any
        by_span = {}
        by_doc = defaultdict(set)
        by_collection = set()
        for doc in system:
            for ann in doc.annotations:
                key = self._get_key(ann.candidates[0])
                by_span[ann] = key
                by_doc[ann.docid].add(key)
                by_collection.add(key)

        by_doc.default_factory = None

        out = []
        for doc in gold:
            for ann in doc.annotations:
                if len(ann.candidates) <= 1:
                    out.append(str(ann))
                    continue

                keys = [self._get_key(cand) for cand in ann.candidates]
                try:
                    matched = keys.index(by_span[ann.span])
                except (KeyError, IndexError):
                    # span not annotated or candidate not matched
                    try:
                        doc_keys = by_doc[ann.docid]
                    except KeyError:
                        # no system annotations in this doc
                        doc_keys = set()
                    collection_match = None
                    for i, key in enumerate(keys):
                        if key in doc_keys:
                            matched = i
                            break
                        if collection_match is None and key in by_collection:
                            collection_match = i
                    else:
                        if collection_match is None:
                            matched = 0
                        else:
                            matched = collection_match

                ann.candidates = [ann.candidates[matched]]
                out.append(str(ann))

        return '\n'.join(out)

    @classmethod
    def add_arguments(cls, p):
        p.add_argument('-f', '--fields', default='eid',
                       help='Comma-delimited list of fields to match '
                            'candidates at the same span between system and '
                            'gold. "*" will require match on all fields; '
                            'default is "eid".')
        p.add_argument('-g', '--gold', required=True,
                       help='Path to gold standard annotations')
        p.add_argument('system', metavar='FILE',
                       help='Path to system annotations')
        p.set_defaults(cls=cls)
        return p


class WeightsForHierarchy(object):
    """Translate a hierarchy of types into a sparse matrix of type-pair weights

    Input is a JSON object mapping parents to children in the hierarchy.
    Output is a three-column TSV with:

        * gold type
        * system type
        * weight

    The weights are assigned such that where the system type is an ancestor of
    the gold type with d edges between them, it will score (decay ** d).
    """

    def __init__(self, hierarchy_json, decay=0.5):
        self.hierarchy_json = hierarchy_json
        self.decay = decay
        if decay > 1.0 or decay < 0:
            raise ValueError('Decay must be greater than 0 and at most 1')

    def __call__(self):
        if self.hierarchy_json.startswith('{'):
            hierarchy = json.loads(self.hierarchy_json)
        else:
            with open(self.hierarchy_json) as f:
                hierarchy = json.load(f)

        out = []
        for parent, children in hierarchy.items():
            self._descend(out, hierarchy, parent, children, self.decay)
        # XXX: returning this all as a string rather than writing to a stream
        #      is yuck!
        return '\n'.join('%s\t%s\t%f' % tup for tup in out)

    def _descend(self, out, hierarchy, gold, children, weight):
        for child in children:
            out.append((gold, child, weight))
            self._descend(out, hierarchy, gold, hierarchy.get(child, ()),
                          weight * self.decay)

    @classmethod
    def add_arguments(cls, p):
        p.add_argument('-d', '--decay', default=0.5, type=float,
                       help='Decay value for systems selecting an ancestor '
                       'of the gold type')
        p.add_argument('hierarchy_json', metavar='FILE',
                       help='Path to hierarchy JSON')
        p.set_defaults(cls=cls)
        return p
